/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <xc.h>
#include <plib.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include "uartBuffer.h"
#include "fatFS/ff.h"
#include "Hashes.h"
#include <stdlib.h>
#include <string.h>

#pragma config IESO     = OFF      // Internal/External Switch-over
#pragma config FNOSC    = PRIPLL   // Oscillator Selection
#pragma config FPLLIDIV = DIV_2    // PLL Input Divider (PIC32 Starter Kit: use divide by 2 only)
#pragma config FPLLMUL  = MUL_20   // PLL Multiplier
#pragma config FPLLODIV = DIV_1    // PLL Output Divider
#pragma config FPBDIV   = DIV_1    // Peripheral Clock divisor
#pragma config FWDTEN   = OFF      // Watchdog Timer
#pragma config WDTPS    = PS1      // Watchdog Timer Postscale
#pragma config FCKSM    = CSDCMD   // Clock Switching & Fail Safe Clock Monitor
#pragma config OSCIOFNC = OFF      // CLKO Enable
#pragma config POSCMOD  = XT       // Primary Oscillator
#pragma config FSOSCEN  = ON       // Secondary Oscillator Enable
#pragma config CP       = OFF      // Code Protect
#pragma config BWP      = OFF      // Boot Flash Write Protect
#pragma config PWP      = OFF      // Program Flash Write Protect
#pragma config ICESEL   = ICS_PGx2 // ICE/ICD Comm Channel Select
#pragma config DEBUG    = OFF      // Debugger
#pragma config FMIIEN = OFF
#pragma config FETHIO = ON

#define GetSystemClock() ( 80000000ul )
#define GetInstructionClock() GetSystemClock()
#define GetPeripheralClock() GetSystemClock()

static FATFS fs;

DWORD get_fattime( void ) {
   return 0;
}

static void runMD5Calc( const char *fileName );
static void readTest( const char *fileName );

static int calcFileMD5( const char *fileName,
      unsigned char md5[ static 16 ] );

int main( void ) {
   SYSTEMConfig( GetSystemClock(),
         SYS_CFG_WAIT_STATES | SYS_CFG_PCACHE);

   UARTConfigure( UART2, UART_ENABLE_PINS_TX_RX_ONLY | UART_ENABLE_HIGH_SPEED );
   UARTSetFifoMode( UART2, UART_INTERRUPT_ON_TX_DONE | UART_INTERRUPT_ON_RX_NOT_EMPTY );
   UARTSetDataRate( UART2, GetPeripheralClock(), 115200 );
   UARTEnable( UART2, UART_ENABLE | UART_PERIPHERAL | UART_RX | UART_TX );

   INTSetVectorPriority( INT_VECTOR_UART( UART2 ), INT_PRIORITY_LEVEL_2 );
   INTSetVectorSubPriority( INT_VECTOR_UART( UART2 ), INT_PRIORITY_LEVEL_1 );
   INTClearFlag( INT_SOURCE_UART( UART2 ) );
   INTEnable( INT_SOURCE_UART_RX( UART2 ), INT_ENABLED );

   INTEnableSystemMultiVectoredInt();
   INTEnableInterrupts();

   puts( "UART initialised." );

   struct command cmds[] =
   {
       { "md5", &runMD5Calc },
       { "read", &readTest },
   };
   uartbuf_init( cmds, sizeof( cmds ) / sizeof( cmds[ 0 ] ) );

   if ( !disk_initialize( 0 ) && !f_mount( 0, &fs ) )
      puts( "FAT32 library initialised." );
   else
      puts( "Failed to initialise FAT32 library." );

   for ( ;; )
   {
       uartbuf_runCmdCallbacks();
   }
 }


void _mon_putc( char c ) {
   while ( U2STAbits.UTXBF );

   U2TXREG = c;
}

void _general_exception_handler( unsigned cause, unsigned status )
{
   printf( "Yikes, exception!\n  Cause: %u\n  Status:  %u\n",
         ( cause & 0x1c ) >> 2, ( status & 0x7c ) >> 2 );
   for (;;);
}


static void runMD5Calc( const char *fileName )
{
   unsigned char md5[ 16 ];

   printf( "Starting to calculate MD5 hash of the file '%s' ...\n",
         fileName );

   FRESULT retVal = calcFileMD5( fileName, md5 );
   if ( retVal )
   {
      printf( "%s failed: %d\n", __func__, retVal );
      return;
   }

   printf( "MD5 of the file '%s' is", fileName );
   int i;
   for ( i = 0; i < sizeof( md5 ); ++i )
      printf( " %02x", md5[ i ] );
   puts( "" );
}

static void readTest( const char *fileName )
{
   HASH_SUM md5;
   MD5Initialize( &md5 );

   FIL f;
   FRESULT retVal = f_open( &f, fileName, FA_READ );
   if ( retVal )
   {
      printf( "%s: Could not open file '%s': %d\n", __func__, fileName,
            retVal );
      return;
   }

   printf( "Starting read test using the file '%s' ...\n", fileName );

   unsigned int duration = 0;
   unsigned int numBytesReadTotal = 0;
   unsigned int prevTimestamp = ReadCoreTimer();
   while ( !f_eof( &f ) )
   {
      unsigned char buffer[ 512 ];
      unsigned int numBytesRead;
      if ( retVal = f_read( &f, buffer, sizeof( buffer ), &numBytesRead ) )
      {
         f_close( &f );
         printf( "%s: Read error while reading '%s': %d\n", __func__,
               fileName, retVal );
         return;
      }

      /* Do not include MD5 hashing work in duration: */
      duration += ReadCoreTimer() - prevTimestamp;

      MD5AddData( &md5, buffer, sizeof( buffer ) );

      numBytesReadTotal += numBytesRead;
      prevTimestamp = ReadCoreTimer();
   }

   f_close( &f );

   printf(  "File reading test completed.\n"
            "  Bytes read:        %u\n"
            "  Duration in ms:    %u\n", numBytesReadTotal, duration /
            ( GetSystemClock() / 2000 ) );

   unsigned char md5Result[ 16 ];
   puts( "Computing MD5 hash ..." );
   MD5Calculate( &md5, md5Result );
   puts( "MD5 hash computed. Comparing against pre-computed value ..." );

   size_t fileNameLen = strlen( fileName );
   char *hashFileName = malloc( ( fileNameLen + 4 + 1 ) * sizeof( char ) );
   if ( !hashFileName )
   {
      puts( "Memory allocation failed" );
      return;
   }

   strcpy( hashFileName, fileName );
   strcpy( &hashFileName[ fileNameLen ], ".md5" );

   if ( retVal = f_open( &f, hashFileName, FA_READ ) )
   {
      printf( "%s: Could not open file '%s':%d\n", __func__, fileName,
            retVal );
      free( hashFileName );
      return;
   }
   free( hashFileName );

   unsigned char md5ExpectedResult[ 16 ];
   unsigned int numBytesRead;
   if ( retVal = f_read( &f, md5ExpectedResult, sizeof( md5ExpectedResult ),
            &numBytesRead ) )
   {
      printf( "%s: Read error while reading '%s': %d\n", __func__,
            hashFileName, retVal );
      f_close( &f );
      return;
   }

   f_close( &f );

   if ( !strncmp( md5Result, md5ExpectedResult, sizeof( md5Result ) ) )
      puts( "MD5 hashes match!" );
   else
      puts( "MD5 HASHES DO NOT MATCH!" );

   puts( "Read test completed." );
}


static int calcFileMD5( const char *fileName, unsigned char md5[ static 16 ] )
{
   FIL f;
   FRESULT retVal = f_open( &f, fileName, FA_READ );
   if ( retVal )
      return retVal;

   HASH_SUM hashSum;
   MD5Initialize( &hashSum );

   while ( !f_eof( &f ) )
   {
      unsigned char buffer[ 64 ];
      unsigned int numBytesRead;
      if ( retVal = f_read( &f, buffer, sizeof( buffer ), &numBytesRead ) )
      {
         f_close( &f );
         return retVal;
      }

      MD5AddData( &hashSum, buffer, sizeof( buffer ) );
   }

   f_close( &f );
   MD5Calculate( &hashSum, md5 );

   return 0;
}
