/**
 * \addtogroup uartBuffer
 * @{
 *
 * \file
 */

#ifndef UARTBUFFER_H
#define UARTBUFFER_H

#include <stddef.h>
#include <stdbool.h>

/**
 * \brief Size of UART receive ring buffer
 *
 * This buffer must be large enough to contain the longest command string, the
 * optional argument and any allowed whitespace in between. It must also hold
 * the newline character.
 */
#define UARTBUF_RX_RINGBUFFER_SIZE 40

/** 
 * \brief Maximum argument length, including whitespace
 */
#define UARTBUF_CMD_ARG_LEN 20

/**
 * \brief The UART module that will be used
 */
#define UARTBUF_UART_MODULE 2

/**
 * \brief The UART module RX interrupt priority
 *
 * Only necessary for PIC32.
 */
#define UARTBUF_UART_INT_PRI 2

/** \brief
 * Command that will execute a callback function when recongised
 */
struct command
{
   /** \brief Command string
    *
    * Must not include newline or whitespace.
    */
   const char *command;
   /** 
    * \brief Function that will be called if #command is recongised
    *
    * If there is a string of characters between the recognised string and
    * newline, preceded by whitepace, it will be given as an argument to the
    * #callback function. If the argument is longer than #UARTBUF_CMD_ARG_LEN,
    * the callback will not be executed. 
    *
    * \param arg argument
    */
   void ( *callback )( const char *arg );
};

/** \brief Initialise the UART input buffer and parser
 *
 * This function initialises an RX ring buffer for an UART module, and
 * optionally initialises a submodule that will call procided callbacks when
 * strings/commands are recognised. If non-NULL, the given list of commands
 * will be traversed, and all the command strings will be compared against the
 * contents of the UART receive buffer every time a newline is read.
 *
 * \param commands array of valid \ref command "commands", which triggers
 * callback functions if the command is recongised.
 * \param numCommands number of items in \pn{commands} array
 */
void uartbuf_init( struct command *commands, size_t numCommands );

/**
 * \brief Run command callbaks
 *
 * If a newline has been received since last time this function was called,
 * the contents of the receive buffer will be compared against the registered
 * commands. The first recongised command string will have its callback
 * executed with arguments retrieved from the remaining characters in the
 * buffer.
 *
 * Call this function repeatedly to check whether a newline has entered the
 * buffer.
 */
void uartbuf_runCmdCallbacks();

/**
 * \brief Copy contents of receive buffer into supplied array
 *
 * \param buffer array to copy data into.
 * \param len length of \pn{array}.
 *
 * \return the number of items copied. May be less than \pn{len} if receive
 * ring buffer do not contain enough items.
 */
size_t uartbuf_copyBuffer( char *buffer, size_t len );

#endif // UARTBUFFER_H

/**
 * @}
 */
