/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "mcuc.h"
#include <stdlib.h>

static enum queryType currPackType = Type_INVALID;
enum state
{
   State_idle,
   State_TX,
   State_RX,
   State_awaitingResponse1,
   State_awaitingResponse2,
   State_aborting,
   State_readRequest,
} static state;
static uint32_t currPackLen;
static uint32_t currPackDataPos;
static uint16_t *currPackData;

static void abortRXTX();
static void handleInternalErr();
static void handleInvalidPacket();
static void allocBuffer();
static bool receiveFragment( union packet packet );
static void transmitFragment();

void mcuc_init()
{
   spi_init();

   state = State_idle;
   currPackType = Type_INVALID;
   currPackData = NULL;

#if defined MCUC_slave
   spi_write( mcuc_makeStatusPacket().data );
   //spi_write( 99 );
#endif
}

void mcuc_doSPIWork()
{
#if defined MCUC_slave 
   if ( spi_recvBufEmpty() )
      return;
#endif

   /* If a packet has been missed due to an overflow, ignore incoming packet,
    * abort any RX/TX process and prepare an error packet: */
   if ( spi_overflow() )
   {
      spi_clearOverflow();
      abortRXTX();

      union packet p = { .error = {
         .abort = true,
         .error = Error_SPIOverflow,
          } };
      p.type = PackType_error;
      spi_write( p.data );
      return;
   }

#if defined MCUC_master
   if ( state == State_idle )
      return;

   if ( state == State_TX )
      transmitFragment();
   else if ( state == State_awaitingResponse2 )
      spi_write( mcuc_makeStatusPacket().data );

   while ( spi_recvBufEmpty() );
#endif
   union packet p = { .data = spi_read() };

   if ( state == State_aborting )
   {
      /* If the last action was to abort, the current response is of no value
       * and must discarded (it was sent before the master has had a chance to
       * react to the abort notification): */
      state = State_idle;
      return;
   }

   /* The packet is a part of a longer transmission: */
   if ( state == State_RX || state == State_readRequest )
   {
      if ( currPackType == Type_INVALID )
      {
         /* This should not happen */
         handleInternalErr();
         return;
      }

      if ( !receiveFragment( p ) )
         /* If there more fragments to be received, send another status
          * packet: */
         spi_write( mcuc_makeStatusPacket().data );
      return;
   }

   /* If the packet is not part of a tranmission, its fields can be inspected
    * */

   if ( p.type == PackType_error )
   {
      if ( p.error.abort )
         abortRXTX();
      mcuc_handleError( p.error );
      spi_write( mcuc_makeStatusPacket().data );
      return;
   }

   if ( state == State_awaitingResponse2 && p.type != PackType_header )
   {
      handleInvalidPacket();
      return;
   }


   if ( p.type == PackType_status )
   {
      mcuc_handleStatus( p );

#if defined MCUC_slave
      /* Send next packet: */
      if ( state == State_TX )
         transmitFragment();
      else
         spi_write( mcuc_makeStatusPacket().data );
#endif

      if ( state == State_awaitingResponse1 )
      {
         state = State_awaitingResponse2;
      }

      return;
   }

   /* Packet is a header packet */

   /* If request is an abort request, abort and exit: */
   if ( p.header.type == Type_abort )
   {
      abortRXTX();
      spi_write( mcuc_makeStatusPacket().data );
      return;
   }

   /* If a query is already being processed, reply with an error and abort: */
   if ( state != State_idle && state != State_awaitingResponse2 )
   {
      union packet p = { .error = {
         .abort = true,
         .error = Error_servingOtherQuery,
      } };
      spi_write( p.data );
      return;
   }

   if ( p.header.query )
   {
      /* A response was expected, but a query was received: */
      if ( state == State_awaitingResponse2 )
      {
         handleInvalidPacket();
         return;
      }

      currPackType = p.header.type;

      if ( p.header.length )
      {
         state = State_readRequest;
         currPackLen = p.header.length;
         allocBuffer();
         spi_write( mcuc_makeStatusPacket().data );
      }
      else 
      {
         state = State_TX;
         currPackLen = 0;
         currPackDataPos = 0;
         mcuc_prepareQueryResponse( currPackType, NULL, 0 );
         transmitFragment();
      }
   }
   else
   {
      /* If an unexpected response is received, report an error: */
      if ( p.header.type != currPackType )
      {
         handleInvalidPacket();
         return;
      }

      state = State_RX;

      currPackLen = p.header.length;

      if ( !currPackLen )
      {
         /* A non-query header packet cannot be of zero length. That is
          * meaningless. Report an error: */
         handleInvalidPacket();
         return;
      }
      
      allocBuffer();
      spi_write( mcuc_makeStatusPacket().data );
   }
}

void mcuc_sendPacket( const union packet *packet )
{
   currPackLen = 1;
   if ( packet->type == PackType_header )
   {
      if ( packet->header.query )
         currPackType = packet->header.type;

      currPackLen += packet->header.length;
   }

   allocBuffer();
   if ( currPackData )
   {
      uint32_t i;
      for ( i = 0; i < currPackLen; ++i )
         currPackData[ i ] = packet[ i ].data;

      state = State_TX;
   }
}

bool mcuc_busy()
{
   return state != State_idle;
}


static void abortRXTX()
{
   state = State_aborting;
   currPackType = Type_INVALID;
   if ( currPackData )
      free( currPackData );
   currPackData = NULL;
}

static void handleInternalErr()
{
   abortRXTX();

   union packet p = { .error = {
      .abort = true,
      .error = Error_internal,
      } };
   p.type = PackType_error;
   spi_write( p.data );
}

static void handleInvalidPacket()
{
   state = State_idle;
   currPackType = Type_INVALID;

   union packet p = { .error = {
      .abort = true,
      .error = Error_packetFormat,
   } };
   p.type = PackType_error;
   spi_write( p.data );
}

static void allocBuffer()
{
   if ( currPackData )
      free( currPackData );

   currPackDataPos = 0;
   currPackData = malloc( currPackLen * sizeof( uint16_t ) );
   if ( !currPackData )
   {
      abortRXTX();

      union packet p = { .error = {
         .abort = true,
         .error = Error_malloc,
      } };
      spi_write( p.data );
   }
}

static bool receiveFragment( union packet packet )
{
   currPackData[ currPackDataPos ] = packet.data;
   ++currPackDataPos;

   if ( currPackDataPos >= currPackLen )
   {
      if ( state == State_readRequest )
      {
         state = State_TX;
         currPackLen = 0;
         currPackDataPos = 0;
         mcuc_prepareQueryResponse( currPackType, currPackData, currPackLen );
         transmitFragment();
         return true;
      }

      if ( mcuc_handleReceivedPacket( currPackType, currPackData,
               currPackLen ) )
         handleInternalErr();
      else
      {
         state = State_idle;
         currPackType = Type_INVALID;
         if ( currPackData )
            free( currPackData );
         currPackData = NULL;
      }

      return true;
   }

   return false;
}

static void transmitFragment()
{
   if ( !currPackData || !currPackLen )
   {
      handleInternalErr();
      return;
   }

   spi_write( currPackData[ currPackDataPos ] );
   ++currPackDataPos;

   if ( currPackDataPos >= currPackLen )
   {
      /* If the transmitted packet was a query, expect to receive a response
       * (got to awaitingResponse state do not change currPackType): */
      if ( ( (union packet *)&currPackData[ 0 ] )->type == PackType_header
               && ( (union packet *)&currPackData[ 0 ] )->header.query )
         state = State_awaitingResponse1;
      else
      {
         state = State_idle;
         currPackType = Type_INVALID;
      }

      if ( currPackData )
         free( currPackData );
      currPackData = NULL;
   }
}
