/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MCUC_PACKET_H
#define MCUC_PACKET_H

#include <stdint.h>
#include "protocol.h"

/**
 * \brief Error packet
 *
 * This packet format must be defined by the user. Note that bit 1:0 are
 * unavailable, and that bit 2 must be called "abort". "abort" will be used to
 * abort a transmission if a reponse is of type error and has this bit set.
 *
 * struct __attribute__(( packed )) errorInfo
 * {
 *    uint16_t               : 2;
 *    uint16_t abort         : 1;
 *    uint16_t               : 13;
 * };
 */

struct __attribute__(( packed )) header
{
   uint16_t               : 2;
   /**
    * \brief Type of packet
    */
   uint16_t type          : 5;
   /**
    * \brief Denotes whether packet is a query or response
    */
   uint16_t query         : 1;
   /**
    * \brief Length of packet in words
    *
    * Length denote the number of additional packet segments, excluding the
    * packet with this header. Maximum length is 255.
    */
   uint16_t length        : 8;
};

enum packetTypes
{
   PackType_header = 0,
   PackType_error  = 1,
   /* 2–3 user-defined */
};

/*
 * \brief Packet format
 *
 * If type is PackType_header (0), the remaining bits of the word are used to
 * form a header. The header can be used to make packets made up of multiple
 * packets that are assembled by the receiver based on the information in the
 * header of the first packet.
 *
 * If type is PackType_error, the remaining bits of the word is used to form a
 * errorInfo object. The format of errorInfo must be defined by the user, but
 * note that bit 2, "abort", is used to abort transmissions if read in a
 * response.
 *
 * Type values 2 and 3 can be used by the user to define special packet
 * formats that do not need a header.
 */
union __attribute__(( packed )) packet
{
   struct __attribute__(( packed ))
   {
      uint16_t type       : 2;
      uint16_t            : 14;
   };
   struct header header;
   struct errorInfo error;
#ifdef MCUC_USER_FORMATS
   MCUC_USER_FORMATS
#endif

   uint16_t data;
};

#endif // MCUC_PACKET_H
