/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "keyboard.h"
#include <Graphics/EditBox.h>
#include <Graphics/Button.h>
#include "wstring.h"
#include "HardwareProfile.h"
#include "tickTimer.h"

static STATICTEXT *entryField;
static XCHAR *textBuffer;
static int textBufferSize;
static int currentPosition;
static BUTTON **buttons;

enum objIDs
{
   OBJID_entryField = 1,
   OBJID_shift,
   OBJID_delete,
   OBJID_symbols,
   OBJID_space,
   OBJID_enter
};

enum caseState
{
   CaseState_lower,
   CaseState_highOnce,
   CaseState_highAlways,
};
static enum caseState caseState = CaseState_lower;
static bool editingFinished;

static XCHAR buttonTextRow0[ 11 ][ 2 ] =
{ L"q", L"w", L"e", L"r", L"t", L"y", L"u", L"i", L"o", L"p", L"å" };
static XCHAR buttonTextRow1[ 11 ][ 2 ] =
{ L"a", L"s", L"d", L"f", L"g", L"h", L"j", L"k", L"l", L"ø", L"æ" };
static XCHAR buttonTextRow2[ 7 ][ 2 ] =
{ L"z", L"x", L"c", L"v", L"b", L"n", L"m" };
static XCHAR buttonTextRow3[ 2 ][ 2 ] =
{ L",", L"." };

static int newLinePositions[ MAX_EDIT_FIELD_LINES ];
static int numEditFieldLines;

#define asciiCaseDiff ( 'a' - 'A' )
#define normalBtnWidth ( GetMaxX() / 11 )
#define normalBtnHeight ( ( GetMaxY() - 100 ) / 4 )
#define normalBtnRadius 3
#define numButtons ( 11 + 11 + 7 + 2 + 5 )

static void changeCase();
static bool addCharacter( XCHAR ch );
static void removeCharacter();
static bool deleteLastWord();
static void deleteAllText();


void keyboard_init( int bufferSize, XCHAR *text )
{
   /* Add room for one more character (the cursor character): */
   textBufferSize = bufferSize + 1;
   textBuffer = malloc( bufferSize * sizeof( XCHAR ) );

   if ( text )
      wstrncpy( textBuffer, text, bufferSize - 1 );
   else
      textBuffer[ 0 ] = L'\0';

   size_t textLen = wstrlen( text );
   if ( textLen >= bufferSize - 1 )
      currentPosition = bufferSize - 1;
   else
      currentPosition = textLen;

   textBuffer[ currentPosition ] = 127;

   entryField = StCreate( OBJID_entryField, 0, 0, GetMaxX(), 100, ST_DRAW,
         textBuffer, 0 );

   buttons = malloc( 36 * sizeof( BUTTON * ) );

   int currX = ( GetMaxX() % 11 ) / 2;
   int currY = 100 + ( GetMaxY() % 4 ) / 2;

   int i;
   for ( i = 0; i < 11; ++i )
   {
      buttons[ i ] = BtnCreate( (WORD)buttonTextRow0[ i ][ 0 ], currX, currY,
            currX + normalBtnWidth, currY + normalBtnHeight, normalBtnRadius,
            BTN_DRAW, NULL, buttonTextRow0[ i ], NULL );
      currX += normalBtnWidth;
   }

   currX = ( GetMaxX() % 11 ) / 2;
   currY += normalBtnHeight;

   for ( i = 0; i < 11; ++i )
   {
      buttons[ 11 + i ] = BtnCreate( (WORD)buttonTextRow1[ i ][ 0 ], currX,
            currY, currX + normalBtnWidth, currY + normalBtnHeight,
            normalBtnRadius, BTN_DRAW, NULL, buttonTextRow1[ i ], NULL );
      currX += normalBtnWidth;
   }

   int specialBtnWidth = ( GetMaxX() - 7 * normalBtnWidth ) / 2;
   currX = ( GetMaxX() - ( 2 * specialBtnWidth + 7 * normalBtnWidth ) ) / 2;
   currY += normalBtnHeight;

   buttons[ 22 ] = BtnCreate( OBJID_shift, currX, currY,
         currX + specialBtnWidth, currY + normalBtnHeight, normalBtnRadius,
         BTN_DRAW, NULL, NULL, NULL );
   currX += specialBtnWidth;

   for ( i = 0; i < 7; ++i )
   {
      buttons[ 23 + i ] = BtnCreate( (WORD)buttonTextRow2[ i ][ 0 ], currX,
            currY, currX + normalBtnWidth, currY + normalBtnHeight,
            normalBtnRadius, BTN_DRAW, NULL, buttonTextRow2[ i ], NULL );
      currX += normalBtnWidth;
   }

   buttons[ 30 ] = BtnCreate( OBJID_delete, currX, currY,
         currX + specialBtnWidth, currY + normalBtnHeight, normalBtnRadius,
         BTN_DRAW, NULL, NULL, NULL );

   currX = 0;
   currY += normalBtnHeight;

   buttons[ 31 ] = BtnCreate( OBJID_symbols, currX, currY,
         currX + specialBtnWidth, currY + normalBtnHeight, normalBtnRadius,
         BTN_DRAW, NULL, NULL, NULL );
   currX += specialBtnWidth;

   buttons[ 32 ] = BtnCreate( (WORD)buttonTextRow3[ 0 ][ 0 ], currX, currY,
         currX + normalBtnWidth, currY + normalBtnHeight, normalBtnRadius,
         BTN_DRAW, NULL, buttonTextRow3[ 0 ], NULL );
   currX += normalBtnWidth;

   buttons[ 33 ] = BtnCreate( OBJID_space, currX, currY,
         currX + GetMaxX() - ( currX * 2 ), currY + normalBtnHeight,
         normalBtnRadius, BTN_DRAW, NULL, NULL, NULL );
   currX += GetMaxX() - ( currX * 2 );

   buttons[ 34 ] = BtnCreate( (WORD)buttonTextRow3[ 1 ][ 0 ], currX, currY,
         currX + normalBtnWidth, currY + normalBtnHeight, normalBtnRadius,
         BTN_DRAW, NULL, buttonTextRow3[ 1 ], NULL );
   currX += normalBtnWidth;

   buttons[ 35 ] = BtnCreate( OBJID_enter, currX, currY,
         currX + specialBtnWidth, currY + normalBtnHeight, normalBtnRadius,
         BTN_DRAW, NULL, NULL, NULL );
}

void keyboard_destroy()
{
   free( textBuffer );
}

WORD keyboard_msgCallback( WORD message, OBJ_HEADER *currentObject,
      GOL_MSG *rawMessage )
{
   WORD id = GetObjID( currentObject );

   static unsigned long keyHoldDuration;
   static unsigned long accumulatedKeyHoldDur;
   static unsigned int requiredHoldTime;
   static struct tick prevPressTimestamp;
   static bool ignoreHeldKey;
   enum
   {
      BackSpace_initialPress,
      BackSpace_deletingWords,
   } static backspaceState;

   if ( id == OBJID_shift )
   {
      if ( message != BTN_MSG_PRESSED )
         return 1;

      if ( caseState == CaseState_lower )
      {
         caseState = CaseState_highOnce;
         changeCase();
      }
      else if ( caseState == CaseState_highOnce )
      {
         caseState = CaseState_highAlways;
      }
      else if ( caseState == CaseState_highAlways )
      {
         caseState = CaseState_lower;
         changeCase();
      }
      return 1;
   }

   if ( id == OBJID_symbols )
   {
      //FIXME
      return 1;
   }

   /* When the button is pressed, start counting the duration of the press: */
   if ( message == BTN_MSG_PRESSED )
   {
      keyHoldDuration = 0;
      accumulatedKeyHoldDur = 0;
      prevPressTimestamp = tickTimer_getTick();
      ignoreHeldKey = false;
      backspaceState = BackSpace_initialPress;

      if ( id == OBJID_delete )
         requiredHoldTime = BACKSPACE_DEL_WORD_HOLD_DELAY;
      else
         requiredHoldTime = KEY_SPECIAL_FUNC_HOLD_DELAY;
   }

   else if ( message == BTN_MSG_STILLPRESSED )
   {
      if ( ignoreHeldKey )
         return 1;

      keyHoldDuration = tickTimer_diff( &prevPressTimestamp );

      if ( keyHoldDuration >= requiredHoldTime )
      {
         if ( id == OBJID_delete )
         {
            /* If backspace has been pressed for a given delay, start deleting
             * words instead of characters: */
            if ( backspaceState == BackSpace_initialPress )
            {
               backspaceState = BackSpace_deletingWords;
               requiredHoldTime = BACKSPACE_DEL_NEXT_WORD_DELAY;
            }
            /* If backspace has been pressed for an even longer delay, delete
             * everything: */
            else if ( backspaceState == BackSpace_deletingWords )
            {
               accumulatedKeyHoldDur += keyHoldDuration;
               if ( accumulatedKeyHoldDur >=
                     BACKSPACE_DEL_EVERYTHING_HOLD_DELAY )
               {
                  deleteAllText();
                  ignoreHeldKey = true;
                  return 1;
               }
            }

            keyHoldDuration = 0;
            prevPressTimestamp = tickTimer_getTick();

            /* If there is no more text to remove, stop: */
            if ( deleteLastWord() )
               return 1;
         }
         else
         {
            //FIXME: Handle key secondary function. Probably needs a state.
            puts( "Keyboard debug: Key secondary function" );
            ignoreHeldKey = true;
            return 1;
         }
      }
      else
         return 1;
   }

   else if ( message != BTN_MSG_PRESSED )
      return 1;

   if ( id == OBJID_delete )
   {
      removeCharacter();
      return 1;
   }

   if ( id == OBJID_enter )
   {
      editingFinished = true;
      // FIXME:
      return 1;
   }

   switch ( id )
   {
      case OBJID_space:
         if ( addCharacter( L' ' ) )
            return 1;
         break;

      default:
         /* There should be no other GOL objects than the ones created by the
          * keyboard, so the remaining IDs can safely be assumed to belong to
          * one of the buttons: */
         if ( addCharacter( BtnGetText( currentObject )[ 0 ] ) )
            return 1;
         break;
   }

   if ( caseState == CaseState_highOnce )
   {
      caseState = CaseState_lower;
      changeCase();
   }

   GOLRedraw( entryField );
   return 1;
}

static void changeCase()
{
   int i;
   for ( i = 0; i < 11; ++i )
      if ( caseState == CaseState_lower )
         buttonTextRow0[ i ][ 0 ] += asciiCaseDiff;
      else
         buttonTextRow0[ i ][ 0 ] -= asciiCaseDiff;

   for ( i = 0; i < 11; ++i )
      if ( caseState == CaseState_lower )
         buttonTextRow1[ i ][ 0 ] += asciiCaseDiff;
      else
         buttonTextRow1[ i ][ 0 ] -= asciiCaseDiff;

   for ( i = 0; i < 7; ++i )
      if ( caseState == CaseState_lower )
         buttonTextRow2[ i ][ 0 ] += asciiCaseDiff;
      else
         buttonTextRow2[ i ][ 0 ] -= asciiCaseDiff;

   if ( caseState == CaseState_lower )
   {
      buttonTextRow3[ 0 ][ 0 ] = L',';
      buttonTextRow3[ 1 ][ 0 ] = L'.';
   }
   else
   {
      buttonTextRow3[ 0 ][ 0 ] = L';';
      buttonTextRow3[ 1 ][ 0 ] = L':';
   }

   for ( i = 0; i < numButtons; ++i )
      GOLRedraw( buttons[ i ] );
}

static bool addCharacter( XCHAR ch )
{
   if ( currentPosition >= textBufferSize - 2 )
      return true;

   textBuffer[ currentPosition ] = ch;
   textBuffer[ ++currentPosition ] = 127;
   textBuffer[ currentPosition + 1 ] = L'\0';

   /* See if text can fit within the width of the display: */
   if ( GetTextWidth( &textBuffer[ newLinePositions[ numEditFieldLines ] ],
            defaultGOLScheme.pFont ) >= GetMaxX() )
   {
      if ( numEditFieldLines + 1 >= MAX_EDIT_FIELD_LINES )
         return true;

      /* See if the newline fits: */
      if ( currentPosition >= textBufferSize - 3 )
         return true;

      textBuffer[ currentPosition ] = L'\n';
      textBuffer[ ++currentPosition ] = 127;
      textBuffer[ currentPosition + 1 ] = L'\0';

      ++numEditFieldLines;
      newLinePositions[ numEditFieldLines ] = currentPosition;
   }

   return false;
}

static void removeCharacter()
{
   if ( currentPosition == 0 )
      return;

   /* currentPosition has cursor character, so go to the character before it:
    * */
   if ( textBuffer[ currentPosition - 1 ] == L'\n' )
   {
      --numEditFieldLines;
      --currentPosition;
   }

   textBuffer[ currentPosition ] = L'\0';
   textBuffer[ --currentPosition ] = 127;

   GOLRedraw( entryField );
}

static bool deleteLastWord()
{
   int i;

   if ( currentPosition == 0 )
      return true;

   /* Find end of previous word (currentPosition has cursor character, so go
    * to the character before it): */
   for ( i = currentPosition - 1; i >= 0 && textBuffer[ i ] == L' '; --i )
      if ( textBuffer[ i ] == L'\n' )
         --numEditFieldLines;

   /* Find start of word: */
   for ( ; i >= 0 && textBuffer[ i ] != L' '; --i )
      if ( textBuffer[ i ] == L'\n' )
         --numEditFieldLines;

   /* Preserve space: */
   ++i;
   currentPosition = i;

   textBuffer[ i ] = 127;
   textBuffer[ i + 1 ] = L'\0';

   GOLRedraw( entryField );
   return false;
}

static void deleteAllText()
{
   currentPosition = 0;
   textBuffer[ 0 ] = 127;
   textBuffer[ 1 ] = L'\0';

   GOLRedraw( entryField );
}
