/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "m29w320FlashDriver.h"
#include <p24Fxxxx.h>
#include <stdbool.h>
#include "HardwareProfile.h"
#ifdef SST_ENABLE_STDOUT_ERRORS
#include <stdio.h>
#endif

#define disableEPMPBypassMode() do { \
      while ( PMCON2bits.BUSY ); \
      PMCON2bits.MSTSEL = 0; \
      } while ( 0 )

#define enableEPMPBypassMode() ( PMCON2bits.MSTSEL = 3 )

/* Declare the flash memory's EDS start address globally, as required for EDS
 * access to work: */
static __eds__ uint16_t __attribute__(( eds, noload,
         address( M29_EPMP_CS_BASE_ADDRESS ) )) flash;


static void rawWrite( uint32_t address, uint16_t data );
static uint16_t rawRead( uint32_t address );
#ifdef SST_ENABLE_STDOUT_ERRORS
static void reportErrorIfOutOfRange( uint32_t address );
#endif
static bool operationDone();

int m29w320_writeWord( uint32_t address, uint16_t data )
{
#ifdef SST_ENABLE_STDOUT_ERRORS
   reportErrorIfOutOfRange( address );
#endif

   disableEPMPBypassMode();

   /* Execute necessary commands for a "word-program operation": */
   rawWrite( 0x555, 0xaa );
   rawWrite( 0x2aa, 0x55 );
   rawWrite( 0x555, 0xa0 );
   rawWrite( address, data );

   while ( !operationDone() );

   /* Perform two successive reads to verify that the data was indeed written
    * (as recommended by the data sheet): */
   bool writeOK = rawRead( address ) == data && rawRead( address ) == data;

   enableEPMPBypassMode();

   return !writeOK;
}

int m29w320_writeBytes( uint32_t address, const uint8_t *data, size_t len )
{
#ifdef SST_ENABLE_STDOUT_ERRORS
   reportErrorIfOutOfRange( address );
#endif

   size_t i;
   for ( i = 0; i < len >> 1; ++i )
      if ( m29w320_writeWord( address++, ( (const uint16_t *)data )[ i ] ) )
         return 1;

   /* If the array contains an odd number of bytes, perform a last one-byte
    * write: */
   if ( len & 1 )
      return m29w320_writeWord( address, data[ len - 1 ] );
   return 0;
}

uint16_t m29w320_readWord( uint32_t address )
{
#ifdef SST_ENABLE_STDOUT_ERRORS
   reportErrorIfOutOfRange( address );
#endif

   disableEPMPBypassMode();
   uint16_t data = rawRead( address );
   enableEPMPBypassMode();

   return data;
}

void m29w320_readBytes( uint32_t address, uint8_t *buffer, size_t bufferLen )
{
#ifdef SST_ENABLE_STDOUT_ERRORS
   reportErrorIfOutOfRange( address );
#endif

   size_t i;
   for ( i = 0; i < bufferLen >> 1; ++i )
      ( ( uint16_t *)buffer )[ i ] = m29w320_readWord( address++ );

   /* If the number of requested bytes is an odd number, perform a last
    * one-byte read: */
   if ( bufferLen & 1 )
      buffer[ bufferLen - 1 ] = m29w320_readWord( address );
}

void m29w320_readBytesBlocking( uint32_t address, uint8_t *buffer,
      size_t bufferLen )
{
#ifdef SST_ENABLE_STDOUT_ERRORS
   reportErrorIfOutOfRange( address );
#endif

   disableEPMPBypassMode();

   size_t i;
   for ( i = 0; i < bufferLen >> 1; ++i )
      ( ( uint16_t *)buffer )[ i ] = rawRead( address++ );

   /* If the number of requested bytes is an odd number, perform a last
    * one-byte read: */
   if ( bufferLen & 1 )
      buffer[ bufferLen - 1 ] = rawRead( address );

   enableEPMPBypassMode();
}

void m29w320_eraseChip()
{
   disableEPMPBypassMode();

   /* Execute necessary commands for a chip erase operation: */
   rawWrite( 0x555, 0xaa );
   rawWrite( 0x2aa, 0x55 );
   rawWrite( 0x555, 0x80 );
   rawWrite( 0x555, 0xaa );
   rawWrite( 0x2aa, 0x55 );
   rawWrite( 0x555, 0x10 );

   while ( !operationDone() );

   enableEPMPBypassMode();
}

void m29w320_eraseBlock( uint32_t address )
{
#ifdef SST_ENABLE_STDOUT_ERRORS
   reportErrorIfOutOfRange( address );
#endif

   disableEPMPBypassMode();

   /* Execute necessary commands for a block erase operation: */
   rawWrite( 0x555, 0xaa );
   rawWrite( 0x2aa, 0x55 );
   rawWrite( 0x555, 0x80 );
   rawWrite( 0x555, 0xaa );
   rawWrite( 0x2aa, 0x55 );
   rawWrite( address, 0x30 );

   while ( !operationDone() );

   enableEPMPBypassMode();
}

int m29w320_size()
{
   disableEPMPBypassMode();

   /* Execute necessary commands for making chip entering CFI query mode: */
   rawWrite( 0x55, 0x98 );

   int sizePower = rawRead( 0x27 );

   /* Exit CFI query mode: */
   rawWrite( 0x555, 0xf0 );

   enableEPMPBypassMode();

   return sizePower;
}

int m29w320_test()
{
   disableEPMPBypassMode();

   rawWrite( 0x555, 0xaa );
   rawWrite( 0x2aa, 0x55 );
   rawWrite( 0x555, 0x90 );

   int manuID = rawRead( 0 );
   rawWrite( 0x555, 0xf0 );

   rawWrite( 0x555, 0xaa );
   rawWrite( 0x2aa, 0x55 );
   rawWrite( 0x555, 0x90 );

   int deviceID = rawRead( 1 );
   rawWrite( 0x555, 0xf0 );

   enableEPMPBypassMode();

   return !( manuID == 0x20 && deviceID == 0x22ca );
}

static void rawWrite( uint32_t address, uint16_t data )
{
   /* Wait for any previous write actions to complete: */
   while ( PMCON2bits.BUSY );

   *(volatile __eds__ uint16_t *)( &flash + address ) = data;
}

static uint16_t rawRead( uint32_t address )
{
   /* Do a "dummy" read of the address. The actual data has to be retrieved
    * using the PMDIN buffer register: */
   (void)*(volatile __eds__ uint16_t *)( &flash + address );

   /* Wait till read has been performed and data is ready in EPMP in buffer:
    * */
   while ( PMCON2bits.BUSY );

   /* Return data from EPMP in buffer: */
   return PMDIN1;
}

#ifdef SST_ENABLE_STDOUT_ERRORS
static void reportErrorIfOutOfRange( uint32_t address )
{
#ifdef M29W320_FLASH_SIZE
   if ( address >= (uint32_t)M29W320_FLASH_SIZE )
      printf( "Attempted to write outside of flash memory's range"
            " (address %lu)\n", address );
#endif
}
#endif

static bool operationDone()
{
   /* Read the DQ6 line two times. If the line is not toggling, the device is
    * ready for a new operation: */
   return ( rawRead( 0 ) & 0x40 ) == ( rawRead( 0 ) & 0x40 );
}
