/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \addtogroup sst39FlashDriver
 * @{
 *
 * \file
 */

#ifndef _SST39FLASHDRIVER_H
#define _SST39FLASHDRIVER_H

#include <stdint.h>
#include <stddef.h>

/**
 * \brief Flash memory device address
 *
 * This define is required.
 */
#define SST_EPMP_CS_BASE_ADDRESS GFX_EPMP_CS2_BASE_ADDRESS

/**
 * \brief Capacity of flash memory device
 *
 * Not required. If defined, and if SST_ENABLE_STDOUT_ERRORS is defined,
 * attempts to write and write outside the address range of the device will be
 * logged as errors to stdout.
 */
#define SST_FLASH_SIZE 0x40000

/**
 * \brief Allows this driver to report errors to stdout
 */
#define SST_ENABLE_STDOUT_ERRORS

/**
 * \brief Write a word (16 bit) to the flash memory
 *
 * This function will temporarily disable EPMP bypass mode, wait until the
 * EPMP module is ready and write the data. It will then busy-wait until the
 * data is written and check that the data was successfully written by reading
 * the same address twice from the flash.
 *
 * \param address the address to where the data should be stored
 * \param data the data to write
 *
 * \return 0 if the write succeeded, non-zero if the data could not be
 * verified after the write operation
 */
int sst39_writeWord( uint32_t address, uint16_t data );

/**
 * \brief Write a series of bytes to the flash memory
 *
 * This functions does the same as sst39_writeWord(), but it operates on a
 * series of bytes instead of one word. EPMP bypass will be disabled and
 * re-enabled between every write operation. \sa sst39_writeWord()
 *
 * \param address the address to where the data should be stored
 * \param data an array of bytes to write
 * \param len length of the \pn{data} array
 *
 * \return 0 if the write succeeded, non-zero if the data could not be
 * verified after the write operation
 */
int sst39_writeBytes( uint32_t address, const uint8_t *data, size_t len );

/**
 * \brief Read a word (16 bit) from the flash memory
 *
 * This function will temporarily disable EPMP bypass mode, request a read
 * from the flash, wait until the EPMP module is ready and return the data
 * from the EPMP buffer.
 *
 * \param address the address to data that will be read
 *
 * \return the requested data
 */
uint16_t sst39_readWord( uint32_t address );

/**
 * \brief Read a series of bytes from the flash memory
 *
 * This function does the same as sst39_readWord(), but it operates on a
 * series of bytes instead of one word. EPMP bypass will be disabled and
 * re-enabled between every read operation. \sa sst39_readWord()
 *
 * \param address the address to data that will be read
 * \param buffer buffer to store data in
 * \param bufferLen length of \pn{buffer}
 */
void sst39_readBytes( uint32_t address, uint8_t *buffer, size_t bufferLen );

/**
 * \brief Read a series of bytes from the flash memory, not re-enabling EPMP
 * bypass mode between writes
 *
 * This function does the same as sst39_readBytes(), but is only disables and
 * re-enables EPMP bypass mode once. \sa sst39_readBytes()
 *
 * \param address the address to data that will be read
 * \param buffer buffer to store data in
 * \param bufferLen length of \pn{buffer}
 */
void sst39_readBytesBlocking( uint32_t address, uint8_t *buffer,
      size_t bufferLen );

/**
 * \brief Erase everything in the flash memory
 */
void sst39_eraseChip();

/**
 * \brief Erase a given sector in the flash memory
 *
 * \param address address of the sector to erase
 */
void sst39_eraseSector( uint32_t address );

/**
 * \brief Get the size of the device
 *
 * \f$2^n\f$, where \f$n\f$ is the return value of this function, gives number of bytes
 * the device can contain.
 */
int sst39_size();

/**
 * \brief Checks the device's manufacture and device ID
 */
int sst39_test();

#endif // _SST39FLASHDRIVER_H

/**
 * @}
 */
