/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "tickTimer.h"
#include "HardwareProfile.h"
#include "peripheralLibConfig.h"
#include </opt/microchip/xc16/current/support/peripheral_24F/timer.h>

static unsigned int interval;
static struct tick currTick;

/**
 * Callback, with function, argument and information about when to fire
 */
struct callback
{
   /**
    * Callback function, returning nothing, taking one argeument, a void
    * pointer
    */
   void ( *func )( void * );
   /** Argument that will be passed to func() */
   void *arg;
   /**
    * Number of times the callbacks should be called (including the first
    * call)
    */
   int count;
   /** The last time the callback fired. Initially 0. */
   struct tick lastFire;
   /** The interval the callback should be called */
   struct tick interval;
} static callbacks[ TICKTIMER_MAX_CALLBACKS ];

#define tickSpecToMS( tickSpec ) ( \
      (unsigned long)tickSpec.s * 1000 + tickSpec.ms )

#define substractTimeSpec( tickSpec1, tickSpec2 ) ( \
      tickSpecToMS( tickSpec1 ) - tickSpecToMS( tickSpec2 ) )

#define PASTER( a, b ) a ## b
#define CONCAT( a, b ) PASTER( a, b )

#define configTimerInt( config ) ( \
      CONCAT( ConfigIntTimer, TICKTIMER_TIMER_NO )( config ) )

#define configTimer( config, period ) ( \
      CONCAT( OpenTimer, TICKTIMER_TIMER_NO )( config, period ) )

#define clearTimerIntFlag() ( \
      CONCAT( T, CONCAT( TICKTIMER_TIMER_NO, _Clear_Intr_Status_Bit ) ) )

#define timerIntFunc() \
   CONCAT( _T, CONCAT( TICKTIMER_TIMER_NO, Interrupt() ) )


int tickTimer_init( unsigned int intervalMs, int priority )
{
   if ( priority < 0 || priority >= 7 )
      return TickTimerInit_errPriority;

   if ( intervalMs == 0 || intervalMs > 1000 )
      return TickTimerInit_errInterval;

   interval = intervalMs;

   /* Flags for timer 1 is used regardless of which hardware timer is chosen,
    * since the flag values are the same for all timers. */
   configTimerInt( priority | T1_INT_ON );
   configTimer( T1_ON | T1_PS_1_256, ( GetInstructionClock() / 1000 *
            intervalMs ) / 256 );

   return TickTimerInit_OK;
}

struct tick tickTimer_getTick()
{
   return currTick;
}

unsigned long tickTimer_diff( const struct tick *lastTick )
{
   return substractTimeSpec( currTick, (*lastTick) );
}

bool tickTimer_hasPassed( const struct tick *lastTick,
      unsigned long durationMs )
{
   return substractTimeSpec( currTick, (*lastTick) ) >= durationMs;
}

int tickTimer_addCallback( void ( *callback )( void *arg ), void *arg,
      int intervalSecs, int intervalMSecs, int repeats )
{
   int i;
   struct callback *vacantCallBack = NULL;

   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
      if ( !callbacks[ i ].func )
      {
         vacantCallBack = &callbacks[ i ];
         break;
      }

   if ( !vacantCallBack )
      return 1;

   vacantCallBack->func = callback;
   vacantCallBack->arg = arg;
   vacantCallBack->count = 1 + repeats;
   vacantCallBack->interval.s = intervalSecs;
   vacantCallBack->interval.ms = intervalMSecs;
   vacantCallBack->lastFire.s = 0;
   vacantCallBack->lastFire.ms = 0;

   return 0;
}

void tickTimer_removeCallback( void ( *callback )( void *arg ) )
{
   int i;
   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
      if ( callbacks[ i ].func == callback )
      {
         callbacks[ i ].func = NULL;
         return;
      }
}

bool tickTimer_callbackActive( void ( *callback )( void *arg ) )
{
   int i;
   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
      if ( callbacks[ i ].func == callback )
         return true;

   return false;
}

void tickTimer_handleCallbacks()
{
   int i;
   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
   {
      if ( callbacks[ i ].func && substractTimeSpec( currTick,
               callbacks[i].lastFire ) >= tickSpecToMS(
                  callbacks[ i ].interval ) )
      {
         ( *callbacks[ i ].func)( callbacks[ i ].arg );
         callbacks[ i ].lastFire = currTick;

         --callbacks[ i ].count;
         if ( !callbacks[ i ].count )
            callbacks[ i ].func = NULL;
      }
   }
}

void __attribute__(( interrupt, auto_psv )) timerIntFunc()
{
   clearTimerIntFlag();

   currTick.ms += interval;
   if ( currTick.ms >= 1000 )
   {
      ++currTick.s;
      currTick.ms = currTick.ms - 1000;
   }
}
