/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "touchScreen.h"
#include "peripheralLibConfig.h"
#include <ports.h>
#include "ft5x06.h"
#include "i2cDriver.h"
#ifdef TOUCH_ENABLE_STDOUT_DEBUG
#include <stdio.h>
#endif
#include <Graphics/DisplayDriver.h>
#include <stdlib.h>

#ifdef TOUCH_ENABLE_STDOUT_DEBUG
static bool debugEnabled;
#endif
static volatile bool touchDataAwaiting;
static volatile bool waitingForI2C;
static bool enabled;

static int fillGOLMsgWithTouchData( GOL_MSG *msg,
      const struct touchInfo *touchInfo );

#ifdef TOUCH_ENABLE_STDOUT_DEBUG
void touch_setDebugEnabled( bool enabled )
{
   debugEnabled = enabled;
}
#endif

void touch_init()
{
   /* Configure interrupt on INT0 with a fairly low priority (should be lower
    * than I²C interrupt priority): */
   ConfigINT0( INT_ENABLE | FALLING_EDGE_INT | INT_PRI_1 );
   enabled = true;
}

void touch_disable()
{
   CloseINT0();
   enabled = false;
   touchDataAwaiting = false;
   waitingForI2C = false;
}

void touch_enable()
{
   touch_init();
}

bool touch_getTouchData( GOL_MSG *msg )
{
   if ( !enabled )
      return false;

   if ( !msg )
   {
#ifdef TOUCH_ENABLE_STDOUT_DEBUG
      puts( "touch_getTouchData(): msg is NULL!" );
#endif
      return false;
   }

   if ( !touchDataAwaiting )
      return false;

   if ( !waitingForI2C )
   {
      ft5x06_queryTouchInfo( 1 );
      waitingForI2C = true;

      return false;
   }

   if ( i2c_busy() )
      return false;

   touchDataAwaiting = false;
   waitingForI2C = false;

   if ( i2c_state == I2C_STATE_error )
   {
#ifdef TOUCH_ENABLE_STDOUT_ERRORS
      printf( "FT5x06 I2C error: %d\n", i2c_error );
#endif
      return false;
   }

   struct touchInfo touchInfo;

   int rv = ft5x06_retrieveTouchInfo( &touchInfo );
   if ( rv != FT5X06_RETTOUCH_OK )
   {
#ifdef TOUCH_ENABLE_STDOUT_ERRORS
      printf( "FT5x06 retrieve error: %d\n", rv );
#endif
      return false;
   }

   rv = fillGOLMsgWithTouchData( msg, &touchInfo );
#ifdef TOUCH_ENABLE_STDOUT_DEBUG
   if ( rv && debugEnabled )
      printf( "FT5x06 warning: Touch coordinates out of bounds (%d,%d)\n",
            touchInfo.points[ 0 ].x, touchInfo.points[ 0 ].y );
#endif

   return true;
}

static int fillGOLMsgWithTouchData( GOL_MSG *msg,
      const struct touchInfo *touchInfo )
{
   /* Keep the previous coordinates in case of an 'put up' event, for which
    * the FT5x06 does not store coordinates: */
   static int x = 0;
   static int y = 0;
   int prevX = x;
   int prevY = y;

   msg->type = TYPE_TOUCHSCREEN;

   /* If there is at least one touch point (finger) (care only about the first
    * touch point): */
   if ( touchInfo->numPoints )
   {
      x = touchInfo->points[ 0 ].x;
      y = touchInfo->points[ 0 ].y;
   }

   /* Translate FT5x06 event to GOL message event: */
   switch ( touchInfo->points[ 0 ].event )
   {
      case FT5X06_EVENT_contact:
         /* Only if finger has moved more than four pixels in either
          * direction, consider it as a move event: */
         if ( abs( x - prevX ) <= 4 && abs( y - prevY ) <= 4 )
            msg->uiEvent = EVENT_STILLPRESS;
         else
            msg->uiEvent = EVENT_MOVE;
         break;

      case FT5X06_EVENT_putDown:
         msg->uiEvent = EVENT_PRESS;
         break;

      case FT5X06_EVENT_putUp:
         msg->uiEvent = EVENT_RELEASE;
         break;

      case FT5X06_EVENT_invalid:
      default:
         msg->uiEvent = EVENT_INVALID;
         break;
   }

   /* The coordinates from the FT5x06 should always be within the allowed
    * range, but make sure the coordinates are valid nevertheless. The IC may
    * be misconfigured. */
   if ( x < 0 )
      x = 0;
   if ( x > GetMaxX() )
      x = GetMaxX();
   if ( y < 0 )
      y = 0;
   if ( y > GetMaxY() )
      y = GetMaxY();

   /* Store either the new coordinate (in case of 'put down' or 'contact'
    * event), or the previous coordinate (in case of 'put up' event, in which
    * case the position of the finger has not moved): */
   msg->param1 = x;
   msg->param2 = y;

   /* Return non-zero if coordinates are out of range: */
   return x < 0 || x > GetMaxX() || y < 0 || y > GetMaxY();
}

void __attribute__(( interrupt, auto_psv )) _INT0Interrupt()
{
   Int0_Clear_Intr_Status_Bit;
   touchDataAwaiting = true;
}
