/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \addtogroup touchScreen
 * @{
 *
 * \file
 */

#ifndef _TOUCHSCREEN_H
#define _TOUCHSCREEN_H

#include <stdbool.h>
#include <Graphics/GOL.h>

/**
 * \brief Allows this module to report errors to stdout
 */
#define TOUCH_ENABLE_STDOUT_ERRORS

/**
 * \brief Allows this module to write debugging information to stdout
 */
#define TOUCH_ENABLE_STDOUT_DEBUG

/**
 * \brief Enable or disable debugging output to stdout
 *
 * \param enabled enable or disable debug info
 */
#ifdef TOUCH_ENABLE_STDOUT_DEBUG
void touch_setDebugEnabled( bool enabled );
#endif

/**
 * \brief Initialises any touch-specific data
 *
 * \note If the underlying touch driver depends on other drives, for instance
 * SPI or I²C, those driver must be initialised before calling this function.
 *
 * If the touch screen driver uses timers or hardware interrupts, those will
 * be set up in this function.
 */
void touch_init();

/**
 * \brief Temporarily disable the touch screen
 *
 * This function makes it possible to disable the touch screen temporarily.
 * touch_getTouchData() can still be called, but it will return false
 * immediatedly.
 */
void touch_disable();

/**
 * \brief Re-enable the touch screen
 *
 * If any underlying drivers has been disabled, this function will not
 * re-enable them.
 */
void touch_enable();

/**
 * \brief Get new touch data when appropriate and fill a GOL message with
 * touch data
 *
 * Must be called regularly from the main loop. It will only do any work if it
 * is neccessary (too long since last poll interval / an interrupt is
 * pending).
 *
 * \param msg message to fill with touch data. Only valid when function
 * returns true
 *
 * \return true if \pn{msg} is filled with new touch data
 */
bool touch_getTouchData( GOL_MSG *msg );

#endif // _TOUCHSCREEN_H

/**
 * @}
 */
