/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "uartDriver.h"
#include "peripheralLibConfig.h"
#include <uart.h>
#include <libpic30.h>

struct ringBuffer
{
   char buffer[ UART_RX_RINGBUFFER_SIZE ];
   size_t start;
   size_t count;
} volatile static ringBuffer;

#define ringBufIncr( i ) ( i = ( i + 1 ) % UART_RX_RINGBUFFER_SIZE )
#define ARG_LEN ( UART_CMD_ARG_LEN + 1 )

#ifdef UART_ENABLE_INPUT_PARSER
static volatile size_t prevNewlinePos;
static volatile size_t newlinePos;
static volatile bool checkCommands;

struct command *commandList;
size_t commandListCount;

static bool ringBufStrExactCmp( const char *string2 );

void uart_init( struct command *commands, size_t numCommands )
#else
void uart_init()
#endif // UART_ENABLE_INPUT_PARSER
{
   ConfigIntUART2( UART_RX_INT_EN | UART_RX_INT_PR1 );
   OpenUART2( UART_EN | UART_BRGH_FOUR, UART_TX_ENABLE, 35 );

   /* Use second UART module as stdout: */
   __C30_UART = 2;

   ringBuffer.start = 0;
   ringBuffer.count = 0;

#ifdef UART_ENABLE_INPUT_PARSER
   checkCommands = false;
   prevNewlinePos = -1;
   newlinePos = -1;
   commandList = commands;
   commandListCount = numCommands;
#endif // UART_ENABLE_INPUT_PARSER
}

#ifdef UART_ENABLE_INPUT_PARSER
void uart_runCmdCallbacks()
{
   if ( !commandList )
      return;

   if ( !checkCommands )
      return;

   DisableIntU2RX;

   checkCommands = false;

   char ch;
   size_t i, spacePos;
   for ( i = ( prevNewlinePos + 1 ) % UART_RX_RINGBUFFER_SIZE; ;
         ringBufIncr( i ) )
      if ( ringBuffer.buffer[ i ] == ' ' || ringBuffer.buffer[ i ] == '\n' )
         break;

   spacePos = i;
   ringBuffer.buffer[ i ] = '\0';

   for ( i = 0; i < commandListCount; ++i )
      if ( commandList[ i ].command && ringBufStrExactCmp( commandList[ i ]
               .command ) )
         break;

   if ( i == commandListCount )
   {
      EnableIntU2RX;
      return;
   }

   void (* callback)( const char *arg ) = commandList[ i ].callback;

   char arg[ ARG_LEN ] = "";

   if ( spacePos != newlinePos )
   {
      for ( i = 0; i < ARG_LEN; ++i )
      {
         ch = ringBuffer.buffer[ ( spacePos + 1 + i ) %
            UART_RX_RINGBUFFER_SIZE ];
         if ( ch == '\n' )
            break;

         arg[ i ] = ch;
      }
      if ( i == ARG_LEN )
      {
         EnableIntU2RX;
         return;
      }

      arg[ i ] = '\0';
   }

   EnableIntU2RX;

   callback( arg );
}


static bool ringBufStrExactCmp( const char *string2 )
{
   size_t i;
   for ( i = 0; ; ++i )
   {
      char ch = ringBuffer.buffer[ ( prevNewlinePos + 1 + i ) %
         UART_RX_RINGBUFFER_SIZE ];

      if ( ch && string2[ i ] )
      {
         if ( ch != string2[ i ] )
            return false;
      }
      else if ( !ch && !string2[ i ] )
         return true;
      else
         return false;
   }

   return true;
}

#endif // UART_ENABLE_INPUT_PARSER

size_t uart_copyBuffer( char *buffer, size_t len )
{
   if ( !buffer )
      return 0;

   size_t i;
   for ( i = 0; i < len && i < ringBuffer.count; ++i )
      buffer[ i ] = ringBuffer.buffer[ ( ringBuffer.start + i ) %
         UART_RX_RINGBUFFER_SIZE ];

   return i;
}


void __attribute__(( interrupt, auto_psv )) _U2RXInterrupt()
{
   size_t end = ( ringBuffer.start + ringBuffer.count ) %
      UART_RX_RINGBUFFER_SIZE;
   char data = U2RXREG;

   ringBuffer.buffer[ end ] = data;

   if ( ringBuffer.count == UART_RX_RINGBUFFER_SIZE )
      ringBufIncr( ringBuffer.start );
   else
      ++ringBuffer.count;

#ifdef UART_ENABLE_INPUT_PARSER
   if ( data == '\n' )
   {
      prevNewlinePos = newlinePos;
      newlinePos = end;
      checkCommands = true;
   }
#endif

   U2RX_Clear_Intr_Status_Bit;
}
