/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \addtogroup uartDriver
 * @{
 *
 * \file
 */

#ifndef UARTDRIVER_H
#define UARTDRIVER_H

#include <stddef.h>
#include <stdbool.h>

/**
 * \brief Size of UART receive ring buffer
 *
 * This buffer must be large enough to contain the longest command string, the
 * optional argument and any allowed whitespace in between. It must also hold
 * the newline character.
 */
#define UART_RX_RINGBUFFER_SIZE 20

/**
 * \brief Enable the input parser
 */
#define UART_ENABLE_INPUT_PARSER

#ifdef UART_ENABLE_INPUT_PARSER

/** 
 * \brief Maximum argument length, including whitespace
 */
#define UART_CMD_ARG_LEN 10

/** \brief
 * Command that will execute a callback function when recongised
 */
struct command
{
   /** \brief Command string
    *
    * Must not include newline or whitespace.
    */
   const char *command;
   /** 
    * \brief Function that will be called if #command is recongised
    *
    * If there is a string of characters between the recognised string and
    * newline, preceded by whitepace, it will be given as an argument to the
    * #callback function. If the argument is longer than #UART_CMD_ARG_LEN,
    * the callback will not be executed. 
    *
    * \param arg argument
    */
   void ( *callback )( const char *arg );
};

/** \brief Initialise the UART module and redirect stdout to it
 *
 * This function initialises a specific UART module (already pin-assigned)
 * with a hard-coded baud rate of 115,200. A list of commands that will be
 * compared against received data can optionally be given. The list will be
 * traversed, and all the command strings will be compared against the
 * contents of the UART receive buffer every time a newline is read.
 *
 * \param commands array of valid \ref command "commands", which triggers
 * callback functions if the command is recongised.
 * \param numCommands number of items in \pn{commands} array
 */
void uart_init( struct command *commands, size_t numCommands );

#else

/** \brief Initialise the UART module and redirect stdout to it
 *
 * This function initialises a specific UART module (already pin-assigned)
 * with a hard-coded baud rate of 115,200.
 */
void uart_init();

#endif // UART_ENABLE_INPUT_PARSER

#ifdef UART_ENABLE_INPUT_PARSER
/**
 * \brief Run command callbaks
 *
 * If a newline has been received since last time this function was called,
 * the contents of the receive buffer will be compared against the registered
 * commands. The first recongised command string will have its callback
 * executed with arguments retrieved from the remaining characters in the
 * buffer.
 *
 * Call this function repeatedly to check whether a newline has entered the
 * buffer.
 */
void uart_runCmdCallbacks();
#endif // UART_ENABLE_INPUT_PARSER

/**
 * \brief Copy contents of receive buffer into supplied array
 *
 * \param buffer array to copy data into.
 * \param len length of \pn{array}.
 *
 * \return the number of items copied. May be less than \pn{len} if receive
 * ring buffer do not contain enough items.
 */
size_t uart_copyBuffer( char *buffer, size_t len );

#endif // UARTDRIVER_H

/**
 * @}
 */
