/* calendar module
 *
 * Used to load and save calendar data (availability types and appointments)
 * to and from files.
 *
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CALENDAR_H
#define CALENDAR_H

#include "calendar.h"
#include "fatFS/ff.h"
#include "fatFS/fsUtils.h"
#include <string.h>
#include <stdlib.h>

static int openAvailFile( uint16_t id, FIL *file, uint8_t flags );
static int createDirs( calDate date );
static int loadCalEntry( struct calEntry *calEntry, FIL *file );
static int saveCalEntry( const struct calEntry *calEntry, FIL *file );
static int makeDefaultUknownAvail();

int cal_numAvailabilities()
{
   DIR dir;
   FRESULT retVal = f_opendir( &dir, CAL_AVAILABILITY_DIR );
   if ( retVal == FR_NO_PATH )
   {
      /* Create the availability directory if it does not already exist: */
      retVal = f_mkdir( CAL_AVAILABILITY_DIR );
      if ( retVal != FR_OK )
         return -retVal;

      /* Since the directory was just created, it obviously does not contains
       * anything. However, the default 'unknown' availability will be created
       * when cal_loadAvailability() is called. Return therefore a positive
       * count: */
      return 1;
   }

   if ( retVal != FR_OK )
      return -retVal;

   int count = 0;
   for ( ;; )
   {
      /* The IDs are 16 bit long, having filenames five characters long, so no
       * LFN support is needed: */
      FILINFO fileInfo = { .lfname = NULL, .lfsize = 0 };

      retVal = f_readdir( &dir, &fileInfo );
      if ( retVal != FR_OK )
         return -retVal;

      /* All items in the directory has been read. Exit: */
      if ( !fileInfo.fname[ 0 ] )
         break;

      /* Do not count '.' and '..': */
      if ( fileInfo.fname[ 0 ] == '.' )
         continue;

      ++count;
   }

   /* If there are no availabilites, something is obviously wrong since the
    * 'unknown' availability is unremovable. Cheat and present 1. A default
    * 'unknown' availability will be created if it doesn't exist later. */
   if ( !count )
      ++count;

   return count;
}

int cal_highestUnusedAvailID()
{
   DIR dir;
   FRESULT retVal = f_opendir( &dir, CAL_AVAILABILITY_DIR );
   if ( retVal == FR_NO_PATH )
   {
      /* Create the availability directory if it does not already exist: */
      retVal = f_mkdir( CAL_AVAILABILITY_DIR );
      if ( retVal != FR_OK )
         return -retVal;

      /* Since the directory was just created, it obviously does not contains
       * anything. However, the default 'unknown' availability will be created
       * when cal_loadAvailability() is called. The first unused ID will
       * therefore never be lower than 1:
       */
      return 1;
   }

   if ( retVal != FR_OK )
      return -retVal;

   int highestID = 0;
   for ( ;; )
   {
      /* The IDs are 16 bit long, having filenames five characters long, so no
       * LFN support is needed: */
      FILINFO fileInfo = { .lfname = NULL, .lfsize = 0 };

      retVal = f_readdir( &dir, &fileInfo );
      if ( retVal != FR_OK )
         return -retVal;

      /* All items in the directory has been read. Exit: */
      if ( !fileInfo.fname[ 0 ] )
         break;

      /* Do not count '.' and '..': */
      if ( fileInfo.fname[ 0 ] == '.' )
         continue;

      {
         char *endPtr;
         long id = strtol( fileInfo.fname, &endPtr, 10 );
         if ( !( *endPtr == '\0' && id >= 0 && id <= INT16_MAX ) )
            return -100;

         if ( id > highestID )
            highestID = id;
      }
   }

   return highestID + 1;
}

int cal_loadAvailability( struct availability *availability, uint16_t id )
{
   if ( !availability )
      return -102;

   FRESULT retVal;
   FIL f;
   unsigned int len, numBytesRead;

   /* Set all strings to NULL so that no unallocated strings will be attempted
    * to be freed if anything goes wrong: */
   availability->title[ 0 ] = NULL;
   availability->title[ 1 ] = NULL;
   availability->description[ 0 ] = NULL;
   availability->description[ 1 ] = NULL;

   retVal = openAvailFile( id, &f, FA_READ );

   /* If the 'unknown' availability (with the special ID 0) has been removed
    * (which should not happen), create a new based on default values: */
   if ( id == 0 && ( retVal == FR_NO_FILE || retVal == FR_NO_PATH ) )
   {
      if ( ( retVal = makeDefaultUknownAvail() ) != FR_OK )
         return retVal;

      /* Open the newly created 'unknown' availability as normal: */
      if ( ( retVal = openAvailFile( id, &f, FA_READ ) ) != FR_OK )
         return retVal;
   }

   else if ( retVal != FR_OK )
      return retVal;

   /* Read all the fixed size entrie to the availability object: */
   len = 2 * 2 + 3 * 4;
   retVal = f_read( &f, &availability->baseType, len, &numBytesRead );
   if ( retVal || numBytesRead != len )
      goto closeAndAbort;

   /* Read availability titles: */
   if ( !( availability->title[ 0 ] = readStringFromFile( &f ) ) )
      goto destroyCloseAndAbort;
   if ( !( availability->title[ 1 ] = readStringFromFile( &f ) ) )
      goto destroyCloseAndAbort;

   /* Read availability descriptions: */
   if ( !( availability->description[ 0 ] = readStringFromFile( &f ) ) )
      goto destroyCloseAndAbort;
   if ( !( availability->description[ 1 ] = readStringFromFile( &f ) ) )
      goto destroyCloseAndAbort;

   f_close( &f );
   return FR_OK;

destroyCloseAndAbort:
   cal_destroyAvailObject( availability );
closeAndAbort:
   f_close( &f );
   /* Return -103 if too few bytes were read, otherwise an error code: */
   return retVal ? retVal : -103;
}

int cal_loadAvailabilities( struct availability *availabilities, int num,
      int *state )
{
   if ( !availabilities )
      return -101;

   DIR dir;
   FRESULT retVal = f_opendir( &dir, CAL_AVAILABILITY_DIR );
   if ( retVal == FR_NO_PATH )
      /* Create the availability directory if it does not already exist. It
       * will obviously be empty, but the unknown availability will be created
       * and returned. */
      retVal = f_mkdir( CAL_AVAILABILITY_DIR );

   if ( retVal != FR_OK )
      return -retVal;

   int posInDir = 0;
   int numLoaded = 0;
   for ( ; numLoaded < num; ++posInDir )
   {
      /* The IDs are 16 bit long, having filenames five characters long, so no
       * LFN support is needed: */
      FILINFO fileInfo = { .lfname = NULL, .lfsize = 0 };

      retVal = f_readdir( &dir, &fileInfo );
      if ( retVal != FR_OK )
         goto freeAndAbort;

      /* All items in the directory has been read. Exit: */
      if ( !fileInfo.fname[ 0 ] )
         break;

      /* Do not count '.' and '..': */
      if ( fileInfo.fname[ 0 ] == '.' )
         continue;

      /* Start loading availabilities from a certain point: */
      if ( state && posInDir < *state )
         continue;

      {
         char *endPtr;
         long id = strtol( fileInfo.fname, &endPtr, 10 );
         if ( !( *endPtr == '\0' && id >= 0 && id <= INT16_MAX ) )
         {
            retVal = 100;
            goto freeAndAbort;
         }

         retVal = cal_loadAvailability( &availabilities[ numLoaded ], id );
         if ( retVal != FR_OK )
            goto freeAndAbort;

         ++numLoaded;
      }
   }

   /* If there were no availabilities (and the directory search was started
    * from the beginning), make the default 'unknown' availability and load
    * it: */
   if ( !numLoaded && ( !state || ( state && !*state ) ) )
   {
      retVal = cal_loadAvailability( &availabilities[ 0 ], 0 );
      if ( retVal != FR_OK )
         goto freeAndAbort;
   }

   /* Update state variable so that if the function is called again with the
    * same variable, availabilities will be fetched from where it stopped: */
   if ( state )
      *state = posInDir;

   /* If fewer than the request number of availabilities could be loaded,
    * invalidate the remaining entries in the array by giving them an invalid
    * base type ID: */
   for ( ; numLoaded < num; ++numLoaded )
      availabilities[ numLoaded ].baseType = AvailBase_INVALID;

   return FR_OK;

freeAndAbort:
   {
      int i;
      for ( i = 0; i < numLoaded; ++i )
         cal_destroyAvailObject( &availabilities[ i ] );
   }

   return -retVal;
}

int cal_saveAvailability( const struct availability *availability )
{
   FRESULT retVal;
   FIL f;
   unsigned int len, numBytesWritten;

   retVal = openAvailFile( availability->id, &f, FA_WRITE |
         FA_CREATE_ALWAYS );
   if ( retVal != FR_OK )
      return retVal;

   /* Save all the fixed size entrie to the availability object: */
   len = 2 * 2 + 3 * 4;
   retVal = f_write( &f, &availability->baseType, len, &numBytesWritten );
   if ( retVal || numBytesWritten != len )
      goto closeAndAbort;

   /* Write availability titles: */
   len = strlen( availability->title[ 0 ] ) + 1;
   retVal = f_write( &f, availability->title[ 0 ], len, &numBytesWritten );
   if ( retVal || numBytesWritten != len )
      goto closeAndAbort;
   len = strlen( availability->title[ 1 ] ) + 1;
   retVal = f_write( &f, availability->title[ 1 ], len, &numBytesWritten );
   if ( retVal || numBytesWritten != len )
      goto closeAndAbort;

   /* Write availability descriptions: */
   len = strlen( availability->description[ 0 ] ) + 1;
   retVal = f_write( &f, availability->description[ 0 ], len,
         &numBytesWritten );
   if ( retVal || numBytesWritten != len )
      goto closeAndAbort;
   len = strlen( availability->description[ 1 ] ) + 1;
   retVal = f_write( &f, availability->description[ 1 ], len,
         &numBytesWritten );
   if ( retVal || numBytesWritten != len )
      goto closeAndAbort;

   f_close( &f );
   return FR_OK;

closeAndAbort:
   f_close( &f );
   /* Return -1 if too few bytes were read, otherwise an error code: */
   return retVal ? retVal : -1;
}

int cal_deleteAvailability( uint16_t id )
{
   /* Allocate space for a string of the form '/availabilites/DDDDD\0': */
   char *fileName = malloc( ( strlen( CAL_AVAILABILITY_DIR ) + 7 ) *
         sizeof( char ) );
   if ( !fileName )
      return -1;

   sprintf( fileName, "%s/%05u", CAL_AVAILABILITY_DIR, id );

   FRESULT retVal = f_unlink( fileName );
   free( fileName );

   return retVal;
}

int cal_makeNewAvailType()
{
   int id = cal_highestUnusedAvailID();
   if ( id < 1 )
      return id;

   struct availability defaultNewAvail =
   {
      .baseType = AvailBase_unavailable,
      .id = id,
      .bgColour1 = CAL_AVAIL_NEW_DEF_BG1,
      .bgColour2 = CAL_AVAIL_NEW_DEF_BG2,
      .textColour = CAL_AVAIL_NEW_DEF_TXT,
      .title =
      {
         CAL_AVAIL_NEW_DEF_TITLE0,
         CAL_AVAIL_NEW_DEF_TITLE1,
      },
      .description =
      {
         CAL_AVAIL_NEW_DEF_DESC0,
         CAL_AVAIL_NEW_DEF_DESC1,
      }
   };

   FRESULT retVal = cal_saveAvailability( &defaultNewAvail );
   if ( retVal != FR_OK )
      return retVal;

   return id;
}

void cal_destroyAvailObject( struct availability *availability )
{
   if ( !availability )
      return;

   if ( availability->title[ 0 ] )
      free( availability->title[ 0 ] );
   if ( availability->title[ 1 ] )
      free( availability->title[ 1 ] );
   if ( availability->description[ 0 ] )
      free( availability->description[ 0 ] );
   if ( availability->description[ 1 ] )
      free( availability->description[ 1 ] );
}

int cal_loadDay( struct calDay *day, calDate date )
{
   if ( !day )
      return -2;

   size_t calDirLen = strlen( CAL_DIR );
   char *fileName = malloc( ( 1 + calDirLen + 10 ) * sizeof( char ) );
   if ( !fileName )
      return -3;

   sprintf( fileName, "%s/%02d/%02d/%02d", CAL_DIR, date.year, date.mon,
      date.mday );
   //printf(" Filename: %s\n", fileName );

   FIL f;
   FRESULT retVal;

   retVal = f_open( &f, fileName, FA_READ );
   free( fileName );

   if ( retVal != FR_OK )
      return retVal;

   unsigned int numBytesRead;

   retVal = f_read( &f, &day->dayNumber, 1, &numBytesRead );
   if ( retVal != FR_OK )
      goto closeAndReturn;
   else if ( numBytesRead != 1 )
   {
      retVal = -1;
      goto closeAndReturn;
   }

   retVal = f_read( &f, &day->numEntries, 1, &numBytesRead );
   if ( retVal != FR_OK )
      goto closeAndReturn;
   else if ( numBytesRead != 1 )
   {
      retVal = -1;
      goto closeAndReturn;
   }

   day->entries = malloc( day->numEntries * sizeof( struct calEntry ) );
   if ( !day->entries )
   {
      retVal = -5;
      goto closeAndReturn;
   }

   int i;
   for ( i = 0; i < day->numEntries; ++i )
   {
      retVal = loadCalEntry( &day->entries[ i ], &f );
      if ( retVal != FR_OK )
      {
         day->numEntries = i;
         cal_destroyDayObject( day );
         goto closeAndReturn;
      }
   }

closeAndReturn:
   f_close( &f );
   return retVal;
}

int cal_saveDay( const struct calDay *day, calDate date )
{
   if ( !day )
      return -2;

   size_t calDirLen = strlen( CAL_DIR );
   char *fileName = malloc( ( 1 + calDirLen + 10 ) * sizeof( char ) );
   if ( !fileName )
      return -3;

   sprintf( fileName, "%s/%02d/%02d/%02d", CAL_DIR, date.year, date.mon,
      date.mday );
   //printf(" Filename: %s\n", fileName );

   FIL f;
   FRESULT retVal;

   retVal = f_open( &f, fileName, FA_WRITE | FA_CREATE_ALWAYS );

   if ( retVal != FR_OK )
   {
      if ( retVal == FR_NO_PATH )
      {
         if ( !( ( retVal = createDirs( date ) ) == FR_OK &&
                 ( retVal = f_open( &f, fileName, FA_WRITE |
                 FA_CREATE_ALWAYS ) ) == FR_OK ) )
         {
            free( fileName );
            return retVal;
         }
      }
      else
      {
         free( fileName );
         return retVal;
      }
   }
   free( fileName );

   unsigned int numBytesWritten;

   retVal = f_write( &f, &day->dayNumber, 1, &numBytesWritten );
   if ( retVal != FR_OK )
      goto closeAndReturn;
   else if ( numBytesWritten != 1 )
   {
      retVal = -1;
      goto closeAndReturn;
   }

   retVal = f_write( &f, &day->numEntries, 1, &numBytesWritten );
   if ( retVal != FR_OK )
      goto closeAndReturn;
   else if ( numBytesWritten != 1 )
   {
      retVal = -1;
      goto closeAndReturn;
   }

   int i;
   for ( i = 0; i < day->numEntries; ++i )
   {
      retVal = saveCalEntry( &day->entries[ i ], &f );
      if ( retVal != FR_OK )
         goto closeAndReturn;
   }

closeAndReturn:
   f_close( &f );
   return retVal;
}

int cal_addEntry( struct calDay *day, const struct calEntry *calEntry )
{
   if ( !day )
      return -1;

   struct calEntry *tempEntries = realloc( day->entries, ( day->numEntries +
            1 ) * sizeof( struct calEntry ) );
   if ( !tempEntries )
      return -2;

   day->entries = tempEntries;
   struct calEntry *e = &day->entries[ day->numEntries ];
   ++day->numEntries;

   /* Copy whole entry, but invalidate the strings so that freeing them in an
    * eventual clean-up does not cause any problems: */
   *e = *calEntry;
   e->appointment.title[ 0 ] = e->appointment.title[ 1 ] =
      e->appointment.description[ 0 ] = e->appointment.description[ 1 ] = NULL;

   e->appointment.title[ 0 ] = malloc( strlen( calEntry->appointment
            .title[ 0 ] ) * sizeof( char ) );
   if ( !e->appointment.title[ 0 ] )
      goto cleanUpAndExit;
   strcpy( e->appointment.title[ 0 ], calEntry->appointment.title[ 0 ] );

   e->appointment.title[ 1 ] = malloc( strlen( calEntry->appointment
            .title[ 1 ] ) * sizeof( char ) );
   if ( !e->appointment.title[ 1 ] )
      goto cleanUpAndExit;
   strcpy( e->appointment.title[ 1 ], calEntry->appointment.title[ 1 ] );

   e->appointment.description[ 0 ] = malloc( strlen( calEntry->appointment
            .description[ 0 ] ) * sizeof( char ) );
   if ( !e->appointment.description[ 0 ] )
      goto cleanUpAndExit;
   strcpy( e->appointment.description[ 0 ], calEntry->appointment
         .description[ 0 ] );

   e->appointment.description[ 1 ] = malloc( strlen( calEntry->appointment
            .description[ 1 ] ) * sizeof( char ) );
   if ( !e->appointment.description[ 1 ] )
      goto cleanUpAndExit;
   strcpy( e->appointment.description[ 1 ], calEntry->appointment
         .description[ 1 ] );

   return FR_OK;

cleanUpAndExit:
   if ( e->appointment.title[ 0 ] )
      free( e->appointment.title[ 0 ] );
   if ( e->appointment.title[ 1 ] )
      free( e->appointment.title[ 1 ] );
   if ( e->appointment.description[ 0 ] )
      free( e->appointment.description[ 0 ] );
   if ( e->appointment.description[ 1 ] )
      free( e->appointment.description[ 1 ] );
   --day->numEntries;
   day->entries = realloc( day->entries, day->numEntries *
         sizeof( struct calEntry ) );

   return -3;
}

void cal_destroyDayObject( struct calDay *day )
{
   if ( !( day && day->entries ) )
      return;

   uint8_t i;
   for ( i = 0; i < day->numEntries; ++i )
   {
      if ( day->entries[ i ].appointment.title[ 0 ] )
         free( day->entries[ i ].appointment.title[ 0 ] );
      if ( day->entries[ i ].appointment.title[ 1 ] )
         free( day->entries[ i ].appointment.title[ 1 ] );
      if ( day->entries[ i ].appointment.description[ 0 ] )
         free( day->entries[ i ].appointment.description[ 0 ] );
      if ( day->entries[ i ].appointment.description[ 1 ] )
         free( day->entries[ i ].appointment.description[ 1 ] );
   }

   free( day->entries );
}


static int openAvailFile( uint16_t id, FIL *file, uint8_t flags )
{
   FRESULT retVal;
   size_t availDirLen = strlen( CAL_AVAILABILITY_DIR );

   /* Full path is [availDir] + "/DDDDD\0": */
   char *fileName = malloc( ( availDirLen + 7 ) * sizeof( char ) );
   if ( !fileName )
      return -1;

   sprintf( fileName, "%s/%05u", CAL_AVAILABILITY_DIR, id );
   //printf( "file name: '%s'\n", fileName );

   retVal = f_open( file, fileName, flags );

   /* If this is a write operation and the path is not found, create the path:
    * */
   if ( ( flags & FA_WRITE ) && retVal == FR_NO_PATH )
   {
      fileName[  availDirLen ] = '\0';
      retVal = f_mkdir( fileName );
      if ( retVal != FR_OK )
      {
         free( fileName );
         return retVal;
      }

      fileName[ availDirLen ] = '/';
      retVal = f_open( file, fileName, flags );
   }

   free( fileName );
   return retVal;
}

static int createDirs( calDate date )
{
   FRESULT retVal;
   size_t calDirLen = strlen( CAL_DIR );
   /* Create full path name: [calDir] + "/DD/DD\0": */
   char *dirName = malloc( ( calDirLen + 7 ) * sizeof( char ) );
   if ( !dirName )
      return -1;

   //printf( "Date: %lu\n", date.l );
   sprintf( dirName, "%s%c%02d%c%02d", CAL_DIR, '\0', date.year, '\0',
      date.mon );
   //printf( "Creating %s\n", dirName );
   retVal = f_mkdir( dirName );
   if ( retVal != FR_OK && retVal != FR_EXIST )
   {
      free( dirName );
      return retVal;
   }

   dirName[ calDirLen ] = '/';
   //printf( "Creating %s\n", dirName );
   retVal = f_mkdir( dirName );
   if ( retVal != FR_OK && retVal != FR_EXIST )
   {
      free( dirName );
      return retVal;
   }

   dirName[ calDirLen + 3 ] = '/';
   //printf( "Creating %s\n", dirName );
   retVal = f_mkdir( dirName );
   free( dirName );
   return retVal;
}

static int loadCalEntry( struct calEntry *calEntry, FIL *file )
{
   unsigned int numBytesRead;
   unsigned int len;
   FRESULT retVal;
   struct appointment *a = &calEntry->appointment;

   /* Set all strings to NULL so that no unallocated strings will be attempted
    * to be freed if anything goes wrong: */
   a->title[ 0 ] = NULL;
   a->title[ 1 ] = NULL;
   a->description[ 0 ] = NULL;
   a->description[ 1 ] = NULL;

   /* Load start and end timestamps and settings: */
   len = 2 * sizeof( calEntry->start ) + sizeof( struct calEntrySettings );
   retVal = f_read( file, &calEntry->start, len, &numBytesRead );
   if ( retVal != FR_OK )
      return retVal;
   else if ( numBytesRead != len )
      return -1;

   /* Load appointment ID and availability ID: */ 
   len = 4;
   retVal = f_read( file, &a->availabilityID, len, &numBytesRead );
   if ( retVal != FR_OK )
      return retVal;
   else if ( numBytesRead != len )
      return -1;

   /* Load appointment title: */
   if ( !( a->title[ 0 ] = readStringFromFile( file ) ) )
      return -4;
   if ( !( a->title[ 1 ] = readStringFromFile( file ) ) )
      return -4;
   /* Load appointment description: */
   if ( !( a->description[ 0 ] = readStringFromFile( file ) ) )
      return -4;
   if ( !( a->description[ 1 ] = readStringFromFile( file ) ) )
      return -4;

   return FR_OK;
}

static int saveCalEntry( const struct calEntry *calEntry, FIL *file )
{
   unsigned int numBytesWritten;
   unsigned int len;
   FRESULT retVal;
   int i;

   /* Save start and end timstamps and settings: */
   len = 2 * sizeof( calEntry->start ) + sizeof( struct calEntrySettings );
   retVal = f_write( file, &calEntry->start, len, &numBytesWritten );
   if ( retVal != FR_OK )
      return retVal;
   else if ( numBytesWritten != len )
      return -1;

   /* Save appointment ID and availability ID: */ 
   len = 4;
   retVal = f_write( file, &calEntry->appointment.availabilityID, len,
         &numBytesWritten );
   if ( retVal != FR_OK )
      return retVal;
   else if ( numBytesWritten != len )
      return -1;

   /* Save appointment title: */
   for ( i = 0; i < 2; ++i )
   {
      len = strlen( calEntry->appointment.title[ i ] ) + 1;
      retVal = f_write( file, calEntry->appointment.title[ i ], len,
            &numBytesWritten );
      if ( retVal != FR_OK )
         return retVal;
      else if ( numBytesWritten != len )
         return -1;
   }
   
   /* Save appointment description: */
   for ( i = 0; i < 2; ++i )
   {
      len = strlen( calEntry->appointment.description[ i ] ) + 1;
      retVal = f_write( file, calEntry->appointment.description[ i ], len,
            &numBytesWritten );
      if ( retVal != FR_OK )
         return retVal;
      else if ( numBytesWritten != len )
         return -1;
   }

   return FR_OK;
}

static int makeDefaultUknownAvail()
{
   struct availability defaultUnkown =
   {
      .baseType = AvailBase_unknown,
      .id = 0,
      .bgColour1 = CAL_AVAIL_UNKNOWN_DEF_BG1,
      .bgColour2 = CAL_AVAIL_UNKNOWN_DEF_BG2,
      .textColour = CAL_AVAIL_UNKNOWN_DEF_TXT,
      .title =
      {
         CAL_AVAIL_UNKNOWN_DEF_TITLE0,
         CAL_AVAIL_UNKNOWN_DEF_TITLE1,
      },
      .description =
      {
         CAL_AVAIL_UNKNOWN_DEF_DESC0,
         CAL_AVAIL_UNKNOWN_DEF_DESC1,
      },
   };

   return cal_saveAvailability( &defaultUnkown );
}

#endif // CALENDAR_H
