/* calendar module
 *
 * Used to load and save calendar data (availability types and appointments)
 * to and from files.
 *
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdbool.h>
#include <peripheral/rtcc.h>

#define CAL_DIR "/cal"
#define CAL_AVAILABILITY_DIR "/availabilities"

#define CAL_AVAIL_UNKNOWN_DEF_BG1         0x006bf5
#define CAL_AVAIL_UNKNOWN_DEF_BG2         0x001499
#define CAL_AVAIL_UNKNOWN_DEF_TXT         0xffffff
#define CAL_AVAIL_UNKNOWN_DEF_TITLE0      "Unknown"
#define CAL_AVAIL_UNKNOWN_DEF_TITLE1      CAL_AVAIL_UNKNOWN_DEF_TITLE0
#define CAL_AVAIL_UNKNOWN_DEF_DESC0       "(Availability not set)"
#define CAL_AVAIL_UNKNOWN_DEF_DESC1       CAL_AVAIL_UNKNOWN_DEF_DESC0

#define CAL_AVAIL_NEW_DEF_BG1             0xf50000                    
#define CAL_AVAIL_NEW_DEF_BG2             0x8a0707                    
#define CAL_AVAIL_NEW_DEF_TXT             0xffffff                    
#define CAL_AVAIL_NEW_DEF_TITLE0          "Busy"                      
#define CAL_AVAIL_NEW_DEF_TITLE1          CAL_AVAIL_NEW_DEF_TITLE0
#define CAL_AVAIL_NEW_DEF_DESC0           "Please do not disturb"     
#define CAL_AVAIL_NEW_DEF_DESC1           CAL_AVAIL_NEW_DEF_DESC0 

typedef rtccDate calDate;
typedef rtccTime calTime;

struct timestamp
{
   calDate date;
   calTime time;
};

enum availabilityBaseType
{
   AvailBase_unknown       = 0,
   AvailBase_available     = 1,
   AvailBase_unavailable   = 2,
   AvailBase_INVALID       = UINT16_MAX,
};

struct availability
{
   uint16_t baseType;
   uint16_t id;
   uint32_t bgColour1;
   uint32_t bgColour2;
   uint32_t textColour;
   char *title[ 2 ];
   char *description[ 2 ];
};

struct appointment
{
   uint16_t appointmentID;
   uint16_t availabilityID;
   char *title[ 2 ];
   char *description[ 2 ];
};

struct __attribute__(( packed )) calEntrySettings
{
   uint8_t titleVisibleWeb        : 1;
   uint8_t titleVisibleDisplay    : 1;
   uint8_t descVisibleWeb         : 1;
   uint8_t descVisibleDisplay     : 1;
   uint8_t lastsForever           : 1;
   uint8_t RESERVED               : 3;
};

struct calEntry
{
   struct timestamp start;
   struct timestamp end;
   struct calEntrySettings settings;
   struct appointment appointment;
};

struct calDay
{
   uint8_t dayNumber;
   uint8_t numEntries;
   struct calEntry *entries;
};


int cal_numAvailabilities();

int cal_highestUnusedAvailID();

int cal_loadAvailability( struct availability *availability, uint16_t id );

/**
 *
 * \param availabilities array to store availability objects in. If fewer than
 * the requested number of availabilities could be loaded, the unloaded
 * elements in the array will have baseType AvailBase_INVALID
 * \param num the maxmimum number of availabilities that should be fetched
 * \param state if non-NULL, will contain the position from where the search
 * for availabilities should start. Should be set to 0 in order to start
 * search from the beginning. In order to continue the search from a specific
 * location, pass the state as it was updated from a previous call. If a write
 * operation has been performed to the file system between calls, the state
 * should be considered invalid.
 *
 */
int cal_loadAvailabilities( struct availability *availabilities, int num,
      int *state );

int cal_saveAvailability( const struct availability *availability );

int cal_deleteAvailability( uint16_t id );

/* Returns ID of new availability (or an error) */
int cal_makeNewAvailType();

void cal_destroyAvailObject( struct availability *availability );

int cal_loadDay( struct calDay *day, calDate date );

int cal_saveDay( const struct calDay *day, calDate date );

int cal_addEntry( struct calDay *day, const struct calEntry *calEntry );

void cal_destroyDayObject( struct calDay *day );
