/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "netwSettings.h"
#include "fatFS/ff.h"
#include <GenericTypeDefs.h>
#include <stdbool.h>
#include <TCPIP Stack/TCPIP.h>
#include <TCPIP Stack/DHCP.h>

/* Set up the global 'application configuration', used by the TCP/IP stack to
 * hold network settings: */
APP_CONFIG AppConfig = {
   /* Using the 'special' MAC address 00:04:a3:00:00:00 makes the TCP/IP stack
    * use the Ethernet transciever's MAC address: */
   .MyMACAddr = (MAC_ADDR){ .v = { 0, 0x04, 0xa3, 0, 0, 0 } },
   /* 10.11.12.13: */
   .MyIPAddr.Val = 0xd0c0b0a,
   .DefaultIPAddr.Val = 0xd0c0b0a,
   /* 255.255.255.0: */
   .MyMask.Val = 0xffffff,
   .DefaultMask.Val = 0xffffff,
   /* 10.11.12.1: */
   .MyGateway.Val = 0x10c0b0a,
   .PrimaryDNSServer.Val = 0x10c0b0a,
   .SecondaryDNSServer.Val = 0x10c0b0a,

   .Flags.bIsDHCPEnabled = false,
   .Flags.bInConfigMode = false,
};

struct __attribute__(( packed )) networkSettings
{
   DWORD_VAL ip;
   DWORD_VAL subnet;
   DWORD_VAL gw;
   DWORD_VAL dns1;
   DWORD_VAL dns2;
   bool dhcp;
};

int storeNetworkSettingsToFile()
{
   FIL f;
   FRESULT retVal;

   retVal = f_open( &f, NETW_SETTINGS_FILE, FA_WRITE | FA_CREATE_ALWAYS );
   /* If the settings sub directory does not exist, try to create it: */
   if ( retVal == FR_NO_PATH )
   {
      retVal = f_mkdir( SETTINGS_DIR );
      if ( retVal != FR_OK )
         return retVal;

      retVal = f_open( &f, NETW_SETTINGS_FILE, FA_WRITE | FA_CREATE_ALWAYS );
   }
   /* If another error is encountered, abort: */
   if ( retVal != FR_OK )
      return retVal;

   struct networkSettings settings =
   {
      .ip = AppConfig.MyIPAddr,
      .subnet = AppConfig.MyMask,
      .gw = AppConfig.MyGateway,
      .dns1 = AppConfig.PrimaryDNSServer,
      .dns2 = AppConfig.SecondaryDNSServer,
      .dhcp = DHCPIsEnabled( 0 ),
   };
   unsigned int numBytesWritten;
   retVal = f_write( &f, &settings, sizeof( settings ), &numBytesWritten );

   f_close( &f );

   if ( retVal )
      return retVal;
   if ( numBytesWritten != sizeof( settings ) )
      return -1;

   return FR_OK;
}

int loadNetworkSettingsFromFile()
{
   FIL f;
   FRESULT retVal;

   retVal = f_open( &f, NETW_SETTINGS_FILE, FA_READ );
   if ( retVal != FR_OK )
      return retVal;

   struct networkSettings settings;
   unsigned int numBytesRead;
   retVal = f_read( &f, &settings, sizeof( settings ), &numBytesRead );

   f_close( &f );

   if ( retVal )
      return retVal;
   if ( numBytesRead != sizeof( settings ) )
      return -1;

   AppConfig.MyIPAddr = settings.ip;
   AppConfig.DefaultIPAddr = settings.ip;
   AppConfig.MyMask = settings.subnet;
   AppConfig.DefaultMask = settings.subnet;
   AppConfig.MyGateway = settings.gw;
   AppConfig.PrimaryDNSServer = settings.dns1;
   AppConfig.SecondaryDNSServer = settings.dns2;

   AppConfig.Flags.bIsDHCPEnabled = settings.dhcp;
   AppConfig.Flags.bInConfigMode = settings.dhcp;
   if ( settings.dhcp )
      DHCPEnable( 0 );

   return FR_OK;
}

void loadFactoryNetworkSettings()
{
   DHCPDisable( 0 );

   /* 10.11.12.13: */
   AppConfig.MyIPAddr.Val = 0xd0c0b0a;
   AppConfig.DefaultIPAddr.Val = 0xd0c0b0a;
   /* 255.255.255.0: */
   AppConfig.MyMask.Val = 0xffffff;
   AppConfig.DefaultMask.Val = 0xffffff;
   /* 10.11.12.1: */
   AppConfig.MyGateway.Val = 0x10c0b0a;
   AppConfig.PrimaryDNSServer.Val = 0x10c0b0a;
   AppConfig.SecondaryDNSServer.Val = 0x10c0b0a;

   AppConfig.Flags.bIsDHCPEnabled = false;
   AppConfig.Flags.bInConfigMode = false;
}
