/*
 * Set and retrieve name and title from file
 *
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "form.h"
#include <TCPIP Stack/TCPIP.h>
#include <stdio.h>
#include "calendar/calendar.h"
#include <string.h>

enum action
{
   Action_NOT_SET,
   Action_delete,
   Action_edit,
};
enum param
{
   Param_NOT_SET,
   Param_type,
   Param_colour,
   Param_title,
   Param_desc,
};
enum coloursSetBits
{
   bg1_set = 1,
   bg2_set = 2,
   txt_set = 4,
};

struct query
{
   int32_t id;
   uint8_t action;
   uint8_t param;
   union value
   {
      char string[ 60 ];
      struct colour
      {
         uint32_t bg1;
         uint32_t bg2;
         uint32_t txt;
         uint8_t coloursSet;
      } colour;
   } value;
};

enum POSTNames
{
   Name_UNDEFINED    = 0,
   Name_action       = 1,
   Name_id           = 2,
   Name_param        = 4,
   Name_bg1          = 8,
   Name_bg2          = 16,
   Name_txt          = 32,
   Name_value        = 64,
};

static char *convertAvailTypeToString( const struct availability *availType );

static bool allNameValuesparsed( const struct query *query,
      uint8_t namesParsed );

enum handleQuery_rv
{
   Query_OK          = FormStatus_OK,
   Query_IOError     = FormStatus_internalErr,
   Query_unknownID   = FormStatus_errData,
   Query_invalidID   = FormStatus_errData,
   Query_invalidData = FormStatus_errData,
   Query_memError    = FormStatus_internalErr,
};
static int handleQuery( const struct query *query );

void HTTPPrint_getAllAvailTypes()
{
   /* This is the first time this function has been called in the request for
    * this connection. Intialise state variables: */
   if ( !curHTTP.callbackPos )
   {
      /* Empty the work buffer if it has been used by another callback: */
      if ( curHTTP.dynVarWorkBuffer )
         free( curHTTP.dynVarWorkBuffer );

      /* Use the work buffer to store a variable that keep track where in the
       * avilabilities directory the last read was performed, so that the next
       * read can be performed from where it stopped: */
      curHTTP.dynVarWorkBuffer = malloc( sizeof( int ) );
      if ( !curHTTP.dynVarWorkBuffer )
         /* Fail silently: */
         return;

      /* Set the variable to 0 so that the search is performed at the start of
       * the availabilities directory: */
      *( (int *)curHTTP.dynVarWorkBuffer ) = 0;
   }
   else
      curHTTP.callbackPos = 0;

   /* Since the size of an availability type is not known before it is loaded,
    * the attempt to send the data may have to be interrupted and postponed. A
    * minimum of 80 bytes for TX is demanded, and should be enough for most
    * cases. In the rare case it is not, the operation is aborted, and the
    * HTTP server is signaled that the function is not done and need to be run
    * again (when there is more room in the TX buffer).
    */
   if ( TCPIsPutReady( sktHTTP ) < 80 )
   {
      /* Need more buffer space. Indicate that callback is not done: */
      curHTTP.callbackPos = 1;
      return;
   }

   if ( !curHTTP.dynVarWorkBuffer )
      /* Fail silently: */
      return;

   int *currPosInAvailDir = (int *)curHTTP.dynVarWorkBuffer;

   struct availability availability;
   if ( cal_loadAvailabilities( &availability, 1, currPosInAvailDir )
         != FR_OK )
      /* Fail silently: */
      return;

   /* There are noe more availability types to fetch: */
   if ( availability.baseType == AvailBase_INVALID )
      return;

   char *availTypeString = convertAvailTypeToString( &availability );
   cal_destroyAvailObject( &availability );

   if ( !availTypeString )
      /* Fail silently: */
      return;

   size_t strLen = strlen( availTypeString );
   if ( strLen > TCPIsPutReady( sktHTTP ) )
   {
      /* 80 bytes was not enough, what bad luck! Wait till function gets
       * called again, hopefully with enough TX buffer space: */
      free( availTypeString );
      curHTTP.callbackPos = 1;
      return;
   }

   TCPPutArray( sktHTTP, (uint8_t *)availTypeString, strLen );
   free( availTypeString );

   /* If there are availability types left to fetch, tell the HTTP server
    * that this function need to be called again: */
   curHTTP.callbackPos = 1;
}

void HTTPPrint_getNewAvailTypeDefVals()
{
   /* This is the first time this function has been called in the request for
    * this connection. Intialise state variables: */
   if ( !curHTTP.callbackPos )
   {
      int newID = cal_makeNewAvailType();
      if ( newID < 1 )
         /* Fail silently: */
         return;

      curHTTP.callbackPos = newID;
   }

   /* Make sure that there is a reasonable amount of room in the TX buffer: */
   if ( TCPIsPutReady( sktHTTP ) < 80 )
      return;

   struct availability newAvail;
   if ( cal_loadAvailability( &newAvail, curHTTP.callbackPos ) != FR_OK )
      /* Fail silently: */
      return;

   char *availTypeString = convertAvailTypeToString( &newAvail );
   cal_destroyAvailObject( &newAvail );

   if ( !availTypeString )
      return;

   size_t strLen = strlen( availTypeString );
   if ( strLen > TCPIsPutReady( sktHTTP ) )
   {
      /* 80 bytes was not enough, what bad luck! Wait till function gets
       * called again, hopefully with enough TX buffer space: */
      free( availTypeString );
      return;
   }

   TCPPutArray( sktHTTP, (uint8_t *)availTypeString, strLen );
   free( availTypeString );

   /* Done sending all data. Tell HTTP server that there is no need to call
    * this function again: */
   curHTTP.callbackPos = 0;
}


HTTP_IO_RESULT form_handleAvailTypePOST()
{
   struct state
   {
      struct query query;
      uint8_t currName;
      uint8_t namesParsed;
   };

   if ( !curHTTP.formWorkBuffer )
   {
      curHTTP.formWorkBuffer = malloc( sizeof( struct state ) );
      if ( !curHTTP.formWorkBuffer )
      {
         curHTTP.formStatus = FormStatus_internalErr;
         return HTTP_IO_DONE;
      }
   }

   /* Allocate a buffer that can hold data with length up to the longest
    * name/value expected. An additional two bytes are required by the
    * HTTPReadPostName()/HTTPReadPostValue() functions.
    *
    * The longest name/value is new value for description. Limit to 60:
    */
   char buffer[ 60 + 2 ];

   struct state *s = (struct state *)curHTTP.formWorkBuffer;
   uint8_t retVal;

   /*
    * Query format:
    * require:
    *    'action',['delete','edit']
    * value == 'delete':
    *    require:
    *       'id',(integer)
    *    action: delete availType of ID id
    *    returns:
    *       0: OK
    *       1: unknown ID
    *       2: invalid ID (cannot delete ID 0)
    *       3: I/O error
    * value == 'edit':
    *    require:
    *       'id',(integer)
    *       'param',['type','colour','title','desc']
    *       if param == 'colour':
    *          'bg1',(string)
    *          'bg2',(string)
    *          'txt',(string)
    *       else:
    *          'value',(string)
    *    action: edit availType of ID id change $param to $value
    *    returns:
    *       0: OK
    *       1: unknown ID
    *       2: invalid ID (cannot change type of ID 0)
    *       3: I/O error
    *       4: value overflow (ignore asbsurdly long values (strings are
    *          stored with dynamic lengths, but the will be displayed on a
    *          small display)
    *
    * Queries cannot be combined.
    */

   for ( ;; )
   {
      switch ( curHTTP.smPost )
      {
         case POSTState_idle:
            /* Reset query: */
            s->query.action = Action_NOT_SET;
            s->query.param = Param_NOT_SET;
            s->query.id = -1;
            s->query.value.colour.coloursSet = 0;
            s->query.value.string[ 0 ] = '\0';

            s->currName = Name_UNDEFINED;
            s->namesParsed = 0;

            curHTTP.smPost = POSTState_readName;
            /* No break; go directly to next state: */

         case POSTState_readName:
            retVal = HTTPReadPostName( (uint8_t *)buffer, sizeof( buffer ) );

            if ( retVal == HTTP_READ_INCOMPLETE )
               return HTTP_IO_NEED_DATA;
            else if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            if ( !strcmp( buffer, "action" ) )
               s->currName = Name_action;
            else if ( !strcmp( buffer, "id" ) )
               s->currName = Name_id;
            else if ( !strcmp( buffer, "param" ) )
               s->currName = Name_param;
            else if ( !strcmp( buffer, "bg1" ) )
               s->currName = Name_bg1;
            else if ( !strcmp( buffer, "bg2" ) )
               s->currName = Name_bg2;
            else if ( !strcmp( buffer, "txt" ) )
               s->currName = Name_txt;
            else if ( !strcmp( buffer, "value" ) )
               s->currName = Name_value;
            else
            {
               curHTTP.formStatus = FormStatus_unknownName;
               return HTTP_IO_DONE;
            }

            curHTTP.smPost = POSTState_readValue;
            /* No break; go directly to next state: */

         case POSTState_readValue:
            retVal = HTTPReadPostValue( (uint8_t *)buffer, sizeof( buffer ) );

            if ( retVal == HTTP_READ_INCOMPLETE )
               return HTTP_IO_NEED_DATA;

            if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            /* Check for illegal combinations: */
            if ( 
                  /* Abort if a name appears twice: */
                  ( s->namesParsed & s->currName ) ||
                  /* Do not allow combining 'value' with 'bg1', 'bg2' or
                   * 'txt': */
                  ( ( s->namesParsed & Name_value ) && ( s->namesParsed &
                     ( Name_bg1 | Name_bg2 | Name_txt ) ) )
               )
            {
               curHTTP.formStatus = FormStatus_errCombination;
               return HTTP_IO_DONE;
            }

            s->namesParsed |= s->currName;

            switch ( s->currName )
            {
               case Name_action:
                  if ( !strcmp( buffer, "delete" ) )
                     s->query.action = Action_delete;
                  else if ( !strcmp( buffer, "edit" ) )
                     s->query.action = Action_edit;
                  else
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  break;

               case Name_id:
                  {
                     char *endPtr;
                     s->query.id = strtoul( buffer, &endPtr, 10 );
                     if ( *endPtr != '\0' )
                     {
                        curHTTP.formStatus = FormStatus_errData;
                        return HTTP_IO_DONE;
                     }
                  }
                  break;

               case Name_param:
                  if ( !strcmp( buffer, "type" ) )
                     s->query.param = Param_type;
                  else if ( !strcmp( buffer, "colour" ) )
                     s->query.param = Param_colour;
                  else if ( !strcmp( buffer, "title" ) )
                     s->query.param = Param_title;
                  else if ( !strcmp( buffer, "desc" ) )
                     s->query.param = Param_desc;
                  else
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  break;
                  
               case Name_bg1:
                  {
                     char *endPtr;
                     s->query.value.colour.bg1 = strtoul( buffer, &endPtr,
                           16 );
                     if ( *endPtr != '\0' )
                     {
                        curHTTP.formStatus = FormStatus_errData;
                        return HTTP_IO_DONE;
                     }
                     s->query.value.colour.coloursSet |= bg1_set;
                  }
                  break;

               case Name_bg2:
                  {
                     char *endPtr;
                     s->query.value.colour.bg2 = strtoul( buffer, &endPtr,
                           16 );
                     if ( *endPtr != '\0' )
                     {
                        curHTTP.formStatus = FormStatus_errData;
                        return HTTP_IO_DONE;
                     }
                     s->query.value.colour.coloursSet |= bg2_set;
                  }
                  break;

               case Name_txt:
                  {
                     char *endPtr;
                     s->query.value.colour.txt = strtoul( buffer, &endPtr,
                           16 );
                     if ( *endPtr != '\0' )
                     {
                        curHTTP.formStatus = FormStatus_errData;
                        return HTTP_IO_DONE;
                     }
                     s->query.value.colour.coloursSet |= txt_set;
                  }
                  break;

               case Name_value:
                  strcpy( s->query.value.string, buffer );
                  break;

               default:
                  break;
            }

            if ( allNameValuesparsed( &s->query, s->namesParsed ) )
            {
               curHTTP.formStatus = handleQuery( &s->query );
               return HTTP_IO_DONE;
            }

            /* Unless parsing is done or any errors have occurred, continue
             * reading next name: */
            curHTTP.smPost = POSTState_readName;
            continue;

         default:
            break;
      }
   }

   /* The function should never reach this point. */
   curHTTP.formStatus = FormStatus_internalErr;
   return HTTP_IO_DONE;
}

static char *convertAvailTypeToString( const struct availability *availType )
{
   if ( !availType )
      return NULL;

   size_t titleLen = strlen( availType->title[ 0 ] );
   size_t descLen = strlen( availType->description[ 0 ] );

   /* Format: ddddd\fddddd\fxxxxxx\fxxxxxx\fxxxxxx\fs\fs\r\n\0 */
   char *string = malloc( ( 33 + titleLen + 1 + descLen + 3 ) *
         sizeof ( char ) );
   if ( !string )
      return NULL;

   /* There is currently no support for handling a second language in the web
    * UI, so only the first of the two languages will be sent: */
   sprintf( string, "%u\f%u\f%06x\f%06x\f%06x\f%s\f%s\r\n", availType->id,
         availType->baseType, availType->bgColour1, availType->bgColour2,
         availType->textColour, availType->title[ 0 ],
         availType->description[ 0 ] );

   return string;
}

static int handleQuery( const struct query *query )
{
   //printf( "Handling query:\n"
   //      "  id: %d\n"
   //      "  action: %d\n"
   //      "  param: %d\n", query->id, query->action, query->param );
   //if ( query->param == Param_colour )
   //   printf( "  colours:\n"
   //         "    bg1: %06x\n"
   //         "    bg2: %06x\n"
   //         "    txt: %06x\n", query->value.colour.bg1,
   //         query->value.colour.bg2, query->value.colour.txt );
   //else
   //   printf( "  value: %s\n", query->value.string );

   if ( query->action == Action_delete )
   {
      if ( cal_deleteAvailability( query->id ) != FR_OK )
         return Query_IOError;
   }

   else if ( query->action == Action_edit )
   {
      struct availability a;
      FRESULT retVal = cal_loadAvailability( &a, query->id );
      if ( retVal != FR_OK )
         return Query_IOError;

      if ( query->param == Param_type )
      {
         char *endPtr;
         a.baseType = strtoul( query->value.string, &endPtr, 10 );
         if ( *endPtr != '\0' )
         {
            cal_destroyAvailObject( &a );
            return Query_invalidData;
         }
      }
      else if ( query->param == Param_colour )
      {
         a.bgColour1 = query->value.colour.bg1;
         a.bgColour2 = query->value.colour.bg2;
         a.textColour = query->value.colour.txt;
      }
      else if ( query->param == Param_title )
      {
         /* Currently only support for one language (ignore second title): */
         char *tempTitle = realloc( a.title[ 0 ], (
               strlen( query->value.string ) + 1 ) * sizeof( char ) );
         if ( !tempTitle )
         {
            cal_destroyAvailObject( &a );
            return Query_memError;
         }

         strcpy( tempTitle, query->value.string );
         a.title[ 0 ] = tempTitle;
      }
      else if ( query->param == Param_desc )
      {
         /* Currently only support for one language (ignore second
          * description): */
         char *tempDesc = realloc( a.description[ 0 ], (
               strlen( query->value.string ) + 1 ) * sizeof( char ) );
         if ( !tempDesc )
         {
            cal_destroyAvailObject( &a );
            return Query_memError;
         }

         strcpy( tempDesc, query->value.string );
         a.description[ 0 ] = tempDesc;
      }
         
      retVal = cal_saveAvailability( &a );
      cal_destroyAvailObject( &a );

      if ( retVal != FR_OK )
         return Query_IOError;
   }

   return Query_OK;
}

static bool allNameValuesparsed( const struct query *query,
      uint8_t namesParsed )
{
   /* If possible, build the requirements that will be used to
    * check whether all the names/values have been parsed (and if
    * so, handle the query and exit): */

   /* 'action' and 'id' is always required: */
   uint8_t requiredNames = Name_action | Name_id;

   /* In order to further build requirements 'action' and 'id' must be parsed:
    * */
   if ( ( namesParsed & requiredNames ) != requiredNames )
      return false;

   if ( query->action == Action_edit )
   {
      requiredNames |= Name_param;
      /* In order to further build requirements 'param' must be parsed: */
      if ( !( namesParsed & Name_param ) )
         return false;

      /* If 'param' == 'colour', require three colour name/value pairs. If
       * not, require one 'value' name/value pair: */
      if ( query->param == Param_colour )
         requiredNames |= Name_bg1 | Name_bg2 | Name_txt;
      else
         requiredNames |= Name_value;
   }

   return namesParsed == requiredNames;
}
