/*
 * Set and retrieve name and title from file
 *
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "form.h"
#include <TCPIP Stack/TCPIP.h>
#include <stdio.h>
#include "bio.h"
#include <string.h>

void HTTPPrint_name()
{
   char *name = (char *)curHTTP.dynVarWorkBuffer;

   if ( !curHTTP.callbackPos )
   {
      /* Empty the work buffer if it has been used by another callback: */
      if ( curHTTP.dynVarWorkBuffer )
         free( curHTTP.dynVarWorkBuffer );

      name = bio_name();
      if ( !name )
      {
         TCPPutString( sktHTTP, (unsigned char *)"(NOT SET)" );
         return;
      }

      curHTTP.callbackPos = strlen( name );
   }

   WORD toSend = TCPIsPutReady( sktHTTP );
   if ( curHTTP.callbackPos < toSend )
      toSend = curHTTP.callbackPos;

   TCPPutArray( sktHTTP, (unsigned char *)&name[ strlen( name ) -
         curHTTP.callbackPos ], toSend );

   curHTTP.callbackPos -= toSend;
}

void HTTPPrint_title()
{
   char *title = (char *)curHTTP.dynVarWorkBuffer;

   if ( !curHTTP.callbackPos )
   {
      /* Empty the work buffer if it has been used by another callback: */
      if ( curHTTP.dynVarWorkBuffer )
         free( curHTTP.dynVarWorkBuffer );

      title = bio_title();
      if ( !title )
      {
         TCPPutString( sktHTTP, (unsigned char *)"(NOT SET)" );
         return;
      }

      curHTTP.callbackPos = strlen( title );
   }

   WORD toSend = TCPIsPutReady( sktHTTP );
   if ( curHTTP.callbackPos < toSend )
      toSend = curHTTP.callbackPos;

   TCPPutArray( sktHTTP, (unsigned char *)&title[ strlen( title ) -
         curHTTP.callbackPos ], toSend );

   curHTTP.callbackPos -= toSend;
}

HTTP_IO_RESULT form_handleBioPOST()
{
   enum options
   {
      opt_NONE    = 0,
      opt_name    = 1,
      opt_title   = 2,
   };

   struct state
   {
      unsigned char optionsSet;
      unsigned char nextVal;
   };

   if ( !curHTTP.formWorkBuffer )
   {
      curHTTP.formWorkBuffer = malloc( sizeof( struct state ) );
      if ( !curHTTP.formWorkBuffer )
      {
         curHTTP.formStatus = FormStatus_internalErr;
         return HTTP_IO_DONE;
      }
   }

   /* Allocate a buffer that can hold data with length up to the longest
    * name/value expected. An additional two bytes are required by the
    * HTTPReadPostName()/HTTPReadPostValue() functions.
    *
    * Put a maximum name/title length to 80:
    */
   char buffer[ 80 + 2 ];

   struct state *s = (struct state *)curHTTP.formWorkBuffer;
   unsigned char retVal;

   for ( ;; )
      switch ( curHTTP.smPost )
      {
         case POSTState_idle:
            s->optionsSet = 0;
            s->nextVal = opt_NONE;
            curHTTP.smPost = POSTState_readName;
            /* No break; go directly to next state: */

         case POSTState_readName:
            retVal = HTTPReadPostName( (unsigned char *)buffer,
                  sizeof( buffer ) );

            if ( retVal == HTTP_READ_INCOMPLETE )
               return HTTP_IO_NEED_DATA;

            if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            if ( !strcmp( buffer, "name" ) )
               s->nextVal = opt_name;
            else if ( !strcmp( buffer, "title" ) )
               s->nextVal = opt_title;
            else
            {
               curHTTP.formStatus = FormStatus_unknownName;
               return HTTP_IO_DONE;
            }

            curHTTP.smPost = POSTState_readValue;
            /* No break; go directly to next state: */

         case POSTState_readValue:
            retVal = HTTPReadPostValue( (unsigned char *)buffer,
                  sizeof( buffer ) );
            
            if ( retVal == HTTP_READ_INCOMPLETE )
               return HTTP_IO_NEED_DATA;
            
            if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            s->optionsSet |= s->nextVal;

            switch ( s->nextVal )
            {
               case opt_name:
                  /* Do not allow empty names: */
                  if ( !strlen( buffer ) )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  
                  retVal = bio_setName( buffer );
                  if ( retVal != FR_OK )
                  {
                     curHTTP.formStatus = FormStatus_internalErr;
                     return HTTP_IO_DONE;
                  }
                  break;

               case opt_title:
                  retVal = bio_setTitle( buffer );
                  if ( retVal != FR_OK )
                  {
                     curHTTP.formStatus = FormStatus_internalErr;
                     return HTTP_IO_DONE;
                  }
                  break;

               default:
                  break;
            }

            /* Make sure all settings have been read: */
            if ( ( s->optionsSet & opt_name ) && ( s->optionsSet &
                     opt_title ) )
            {
               curHTTP.formStatus = FormStatus_OK;
               return HTTP_IO_DONE;
            }

            curHTTP.smPost = POSTState_readName;
            continue;

         default:
            break;
      
      }

   /* The function should never reach this point. */
   curHTTP.formStatus = FormStatus_internalErr;
   return HTTP_IO_DONE;
}
