/*
 * Set and retrieve name and title from file
 *
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "form.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <TCPIP Stack/DHCP.h>
#include "tickTimer.h"
#include "config/netwSettings.h"

struct networkSettings
{
   IP_ADDR ip;
   IP_ADDR subnet;
   IP_ADDR gw;
   IP_ADDR dns1;
   IP_ADDR dns2;
   bool persistent;
   bool dhcp;
};

static void setNetworkSettings( void *settings );
static void ipToString( IP_ADDR ipAddr, char *string, size_t strLen );

void HTTPPrint_ipAddr()
{
   char ipAddr[ 16 ];

   ipToString( AppConfig.MyIPAddr, ipAddr, sizeof( ipAddr ) );
   TCPPutString( sktHTTP, (BYTE *)ipAddr );
}

void HTTPPrint_subAddr()
{
   char subnetAddr[ 16 ];

   ipToString( AppConfig.MyMask, subnetAddr, sizeof( subnetAddr ) );
   TCPPutString( sktHTTP, (BYTE *)subnetAddr );
}

void HTTPPrint_gwAddr()
{
   char gwAddr[ 16 ];

   ipToString( AppConfig.MyGateway, gwAddr, sizeof( gwAddr ) );
   TCPPutString( sktHTTP, (BYTE *)gwAddr );
}

void HTTPPrint_dnsAddr1()
{
   char dnsAddr[ 16 ];

   ipToString( AppConfig.PrimaryDNSServer, dnsAddr, sizeof( dnsAddr ) );
   TCPPutString( sktHTTP, (BYTE *)dnsAddr );
}

void HTTPPrint_dnsAddr2()
{
   char dnsAddr[ 16 ];

   ipToString( AppConfig.SecondaryDNSServer, dnsAddr, sizeof( dnsAddr ) );
   TCPPutString( sktHTTP, (BYTE *)dnsAddr );
}

void HTTPPrint_dhcpOn()
{
   TCPPut( sktHTTP, '0' + DHCPIsEnabled( 0 ) );
}

HTTP_IO_RESULT form_handleNetworkSettingsPOST()
{
   enum options
   {
      opt_NONE    = 0,
      opt_temp    = 1,
      opt_dhcp    = 2,
      opt_ip      = 4,
      opt_subnet  = 8,
      opt_gw      = 16,
      opt_dns1    = 32,
      opt_dns2    = 64,
   };

   struct state
   {
      struct networkSettings netw;
      unsigned char optionsSet;
      unsigned char nextVal;
   };

   if ( !curHTTP.formWorkBuffer )
   {
      curHTTP.formWorkBuffer = malloc( sizeof( struct state ) );
      if ( !curHTTP.formWorkBuffer )
      {
         curHTTP.formStatus = FormStatus_internalErr;
         return HTTP_IO_DONE;
      }
   }

   /* Allocate a buffer that can hold data with length up to the longest
    * name/value expected. An additional two bytes are required by the
    * HTTPReadPostName()/HTTPReadPostValue() functions.
    *
    * IP address: xxx.xxx.xxx.xxx (15)
    */
   char buffer[ 15 + 2 ];

   struct state *s = (struct state *)curHTTP.formWorkBuffer;
   unsigned char retVal;

   for ( ;; )
      switch ( curHTTP.smPost )
      {
         case POSTState_idle:
            s->optionsSet = 0;
            s->nextVal = opt_NONE;
            curHTTP.smPost = POSTState_readName;
            /* No break; go directly to next state: */

         case POSTState_readName:
            retVal = HTTPReadPostName( (unsigned char *)buffer,
                  sizeof( buffer ) );

            if ( retVal == HTTP_READ_INCOMPLETE )
            {
               /* Allow all other settings to be skipped if enableDHCP is set,
                * and its value is 'on': */
               if ( ( s->optionsSet | opt_dhcp ) && s->netw.dhcp )
               {
                  curHTTP.formStatus = FormStatus_OK;

                  /* Delay the network settings change a tad so that the HTTP
                   * response can be sent: */
                  tickTimer_addCallback( &setNetworkSettings, &s->netw, 800,
                        0 );
                  return HTTP_IO_DONE;
               }
               else
                  return HTTP_IO_NEED_DATA;
            }

            if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            if ( !strcmp( buffer, "temporary" ) )
               s->nextVal = opt_temp;
            else if ( !strcmp( buffer, "enableDHCP" ) )
               s->nextVal = opt_dhcp;
            else if ( !strcmp( buffer, "ipAddr" ) )
               s->nextVal = opt_ip;
            else if ( !strcmp( buffer, "subnet" ) )
               s->nextVal = opt_subnet;
            else if ( !strcmp( buffer, "gateway" ) )
               s->nextVal = opt_gw;
            else if ( !strcmp( buffer, "primaryDNS" ) )
               s->nextVal = opt_dns1;
            else if ( !strcmp( buffer, "secondaryDNS" ) )
               s->nextVal = opt_dns2;
            else
            {
               curHTTP.formStatus = FormStatus_unknownName;
               return HTTP_IO_DONE;
            }

            curHTTP.smPost = POSTState_readValue;
            /* No break; go directly to next state: */

         case POSTState_readValue:
            retVal = HTTPReadPostValue( (unsigned char *)buffer,
                  sizeof( buffer ) );

            if ( retVal == HTTP_READ_INCOMPLETE )
               return HTTP_IO_NEED_DATA;

            if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            s->optionsSet |= s->nextVal;

            /* Do not allow any empty input: */
            if ( buffer[ 0 ] == '\0' )
            {
               curHTTP.formStatus = FormStatus_errData;
               return HTTP_IO_DONE;
            }

            switch ( s->nextVal )
            {
               case opt_temp:
                  s->netw.persistent = !strcmp( buffer, "no" );
                  break;

               case opt_dhcp:
                  s->netw.dhcp = !strcmp( buffer, "on" );
                  break;

               case opt_ip:
                  if ( !StringToIPAddress( (unsigned char *)buffer,
                           &s->netw.ip ) )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  break;

               case opt_subnet:
                  if ( !StringToIPAddress( (unsigned char *)buffer,
                           &s->netw.subnet ) )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  break;

               case opt_gw:
                  if ( !StringToIPAddress( (unsigned char *)buffer,
                           &s->netw.gw ) )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  break;

               case opt_dns1:
                  if ( !StringToIPAddress( (unsigned char *)buffer,
                           &s->netw.dns1 ) )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  break;

               case opt_dns2:
                  if ( !StringToIPAddress( (unsigned char *)buffer,
                           &s->netw.dns2 ) )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  break;

               default:
                  break;
            }

            /* Ignore the temporary/persistent setting, and make sure all the
             * other settings have been read: */
            if ( ( s->optionsSet & ~opt_temp ) == ( opt_dhcp | opt_ip |
                     opt_subnet | opt_gw | opt_dns1 | opt_dns2 ) )
            {
               curHTTP.formStatus = FormStatus_OK;

               /* Delay the network settings change a tad so that the HTTP
                * response can be sent: */
               tickTimer_addCallback( &setNetworkSettings, &s->netw, 800, 0 );
               return HTTP_IO_DONE;
            }

            curHTTP.smPost = POSTState_readName;
            continue;

         default:
            break;
      }

   /* The function should never reach this point. */
   curHTTP.formStatus = FormStatus_internalErr;
   return HTTP_IO_DONE;
}


static void setNetworkSettings( void *settings )
{
   struct networkSettings *s = (struct networkSettings *)settings;

   AppConfig.Flags.bIsDHCPEnabled = s->dhcp;
   AppConfig.Flags.bInConfigMode = s->dhcp;

   if ( s->dhcp )
      DHCPEnable( 0 );
   else
   {
      DHCPDisable( 0 );

      AppConfig.MyIPAddr = s->ip;
      AppConfig.DefaultIPAddr = s->ip;
      AppConfig.MyMask = s->subnet;
      AppConfig.DefaultMask = s->subnet;
      AppConfig.MyGateway = s->gw;
      AppConfig.PrimaryDNSServer = s->dns1;
      AppConfig.SecondaryDNSServer = s->dns2;
   }
   if ( s->persistent )
   {
      int retVal = storeNetworkSettingsToFile();
      if ( retVal )
         printf( "%s at %u: ERROR: Could not store network settings to file:"
               " %d\n", __func__, __LINE__, retVal );
      else
         puts( "Network settings saved to file" );
   }
}

static void ipToString( IP_ADDR ipAddr, char *string, size_t strLen )
{
   snprintf( string, strLen, "%u.%u.%u.%u", ipAddr.v[ 0 ], ipAddr.v[ 1 ],
         ipAddr.v[ 2 ], ipAddr.v[ 3 ] );
}
