/*
 * Set and retrieve name and title from file
 *
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "form.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <TCPIP Stack/SNTP.h>
#include "config/timeSettings.h"
#include "dateTime.h"

void HTTPPrint_poolServer()
{
   /* This variable is meant to be used to keep track of how many bytes that
    * remain to be sent: */
   if ( !curHTTP.callbackPos )
      curHTTP.callbackPos = strlen( SNTPGetPoolServer() );

   WORD toSend = TCPIsPutReady( sktHTTP );
   if ( curHTTP.callbackPos < toSend )
      toSend = curHTTP.callbackPos;
   
   TCPPutArray( sktHTTP, (unsigned char *)&SNTPGetPoolServer()[ strlen(
            SNTPGetPoolServer() ) - curHTTP.callbackPos ], toSend );

   curHTTP.callbackPos -= toSend;
}

void HTTPPrint_SNTPInterval()
{
   char buffer[ 11 ];
   snprintf( buffer, 11, "%llu", SNTPGetQueryInterval() / TICK_SECOND );
   TCPPutString( sktHTTP, (unsigned char *)buffer );
}

void HTTPPrint_hourOffset()
{
   char buffer[ 12 ];
   snprintf( buffer, 12, "%d", abs( getRTCCOffset() ) );
   TCPPutString( sktHTTP, (unsigned char *)buffer );
}

void HTTPPrint_hourOffsetSign()
{
   TCPPut( sktHTTP, '0' + ( getRTCCOffset() >= 0 ) );
}

void HTTPPrint_date()
{
   char buffer[ 11 ];
   struct tm t = getRTCCDateTime();
   snprintf( buffer, 11, "%u-%u-%u", 1900 + t.tm_year, 1 + t.tm_mon,
         t.tm_mday );
   TCPPutString( sktHTTP, (unsigned char *)buffer );
}

void HTTPPrint_dateArray()
{
   char buffer[ 13 ];
   struct tm t = getRTCCDateTime();
   snprintf( buffer, 13, "%u, %u, %u", 1900 + t.tm_year, 1 + t.tm_mon,
         t.tm_mday );
   TCPPutString( sktHTTP, (unsigned char *)buffer );
}

void HTTPPrint_time()
{
   char buffer[ 6 ];
   struct tm t = getRTCCDateTime();
   snprintf( buffer, 6, "%02u:%02u", t.tm_hour, t.tm_min );
   TCPPutString( sktHTTP, (unsigned char *)buffer );
}

void HTTPPrint_SNTPDisabled()
{
   TCPPut( sktHTTP, '0' + !SNTPEnabled() );
}

HTTP_IO_RESULT form_handleTimeSettingsPOST()
{
   enum options
   {
      opt_NONE          = 0,
      opt_temp          = 1,
      opt_pool          = 2,
      opt_interval      = 4,
      opt_offset        = 8,
      opt_offsetSign    = 16,
      opt_disableSNTP   = 32,
      opt_date          = 64,
      opt_time          = 128,
      opt_updateSNTPNow = 256,
   };

   struct state
   {
      unsigned int optionsSet;
      unsigned int nextVal;
      bool temporaryChanges;
      char poolServer[ NTP_SERVER_MAX_LEN ];
      unsigned long interval;
      int hourOffset;
      bool hourOffsetPositive;
      bool disableSNTP;
      struct tm dateTime;
   };

   if ( !curHTTP.formWorkBuffer )
   {
      curHTTP.formWorkBuffer = malloc( sizeof( struct state ) );
      if ( !curHTTP.formWorkBuffer )
      {
         curHTTP.formStatus = FormStatus_internalErr;
         return HTTP_IO_DONE;
      }
   }

   /* Allocate a buffer that can hold data with length up to the longest
    * name/value expected. An additional two bytes are required by the
    * HTTPReadPostName()/HTTPReadPostValue() functions.
    *
    * SNTP pool server name: Max 30 characters allowed.
    * (Date: YYYY-mm-dd (10))
    */
   char buffer[ 30 + 2 ];

   struct state *s = (struct state *)curHTTP.formWorkBuffer;
   unsigned char retVal;
   unsigned int SNTPOnReqOpts = opt_pool | opt_interval | opt_offset |
      opt_offsetSign;
   unsigned int SNTPOffReqOpts = opt_date | opt_time;

   for ( ;; )
      switch ( curHTTP.smPost )
      {
         case POSTState_idle:
            s->optionsSet = 0;
            s->nextVal = opt_NONE;
            curHTTP.smPost = POSTState_readName;
            /* No break; go directly to next state: */

         case POSTState_readName:
            retVal = HTTPReadPostName( (unsigned char *)buffer,
                  sizeof( buffer ) );

            if ( retVal == HTTP_READ_INCOMPLETE )
            {
               /* If SNTP is temporarily disabled, only care about 'date' and
                * 'time', otherwise only care about 'pool', 'interval',
                * 'offset' and 'offsetSign': */
               if ( ( s->optionsSet & opt_disableSNTP ) &&
                     ( ( s->disableSNTP && ( s->optionsSet & SNTPOffReqOpts ) ==
                         SNTPOffReqOpts ) || ( !s->disableSNTP &&
                            ( s->optionsSet & SNTPOnReqOpts ) ==
                            SNTPOnReqOpts ) ) )
                  break;
               else
                  return HTTP_IO_NEED_DATA;
            }

            if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            if ( !strcmp( buffer, "sntpUpdate" ) )
            {
               SNTPForceUpdate();
               curHTTP.formStatus = FormStatus_OK;
               return HTTP_IO_DONE;
            }
            else if ( !strcmp( buffer, "temporary" ) )
               s->nextVal = opt_temp;
            else if ( !strcmp( buffer, "pool" ) )
               s->nextVal = opt_pool;
            else if ( !strcmp( buffer, "interval" ) )
               s->nextVal = opt_interval;
            else if ( !strcmp( buffer, "offset" ) )
               s->nextVal = opt_offset;
            else if ( !strcmp( buffer, "offsetSign" ) )
               s->nextVal = opt_offsetSign;
            else if ( !strcmp( buffer, "disableSNTP" ) )
               s->nextVal = opt_disableSNTP;
            else if ( !strcmp( buffer, "date" ) )
               s->nextVal = opt_date;
            else if ( !strcmp( buffer, "time" ) )
               s->nextVal = opt_time;
            else
            {
               curHTTP.formStatus = FormStatus_unknownName;
               return HTTP_IO_DONE;
            }

            curHTTP.smPost = POSTState_readValue;
            /* No break; go directly to next state: */

         case POSTState_readValue:
            retVal = HTTPReadPostValue( (unsigned char *)buffer,
                  sizeof( buffer ) );

            if ( retVal == HTTP_READ_INCOMPLETE )
               return HTTP_IO_NEED_DATA;

            if ( retVal == HTTP_READ_TRUNCATED )
            {
               curHTTP.formStatus = FormStatus_overflow;
               return HTTP_IO_DONE;
            }

            s->optionsSet |= s->nextVal;

            /* Do not allow any empty input: */
            if ( buffer[ 0 ] == '\0' )
            {
               curHTTP.formStatus = FormStatus_errData;
               return HTTP_IO_DONE;
            }

            switch ( s->nextVal )
            {
               case opt_temp:
                  s->temporaryChanges = !strcmp( buffer, "yes" );
                  break;

               case opt_pool:
                  *s->poolServer = '\0';
                  strncat( s->poolServer, buffer, sizeof( s->poolServer )
                        - 1 );
                  break;

               case opt_interval:
                  {
                     char *endPtr;
                     s->interval = strtoul( buffer, &endPtr, 10 );
                     if ( *endPtr != '\0' )
                     {
                        curHTTP.formStatus = FormStatus_errData;
                        return HTTP_IO_DONE;
                     }
                  }
                  break;

               case opt_offset:
                  {
                     char *endPtr;
                     s->hourOffset = strtol( buffer, &endPtr, 10 );
                     if ( *endPtr != '\0' )
                     {
                        curHTTP.formStatus = FormStatus_errData;
                        return HTTP_IO_DONE;
                     }
                  }
                  break;

               case opt_offsetSign:
                  s->hourOffsetPositive = !strcmp( buffer, "positive" );
                  break;

               case opt_disableSNTP:
                  s->disableSNTP = !strcmp( buffer, "yes" );
                  break;

               case opt_date:
                  if ( sscanf( buffer, "%u-%u-%u", &s->dateTime.tm_year,
                           &s->dateTime.tm_mon, &s->dateTime.tm_mday ) != 3 )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  s->dateTime.tm_year -= 1900;
                  s->dateTime.tm_mon -= 1;
                  break;

               case opt_time:
                  if ( sscanf( buffer, "%u:%u", &s->dateTime.tm_hour,
                           &s->dateTime.tm_min ) != 2 )
                  {
                     curHTTP.formStatus = FormStatus_errData;
                     return HTTP_IO_DONE;
                  }
                  s->dateTime.tm_sec = 0;
                  break;

               default:
                  break;
            }

            /* If SNTP is temporarily disabled, only care about 'date' and
             * 'time', otherwise only care about 'pool', 'interval',
             * 'offset' and 'offsetSign': */
               if ( ( s->optionsSet & opt_disableSNTP ) &&
                     ( ( s->disableSNTP && ( s->optionsSet & SNTPOffReqOpts )
                         == SNTPOffReqOpts ) || ( !s->disableSNTP && (
                               s->optionsSet & SNTPOnReqOpts ) ==
                            SNTPOnReqOpts ) ) )
            {
               if ( s->disableSNTP )
               {
                  SNTPDisable();
                  setRTCCDateTime( &s->dateTime );
               }
               else
               {
                  SNTPEnable();
                  SNTPSetPoolServer( s->poolServer );
                  SNTPSetQueryInterval( s->interval * TICK_SECOND );
                  if ( !s->hourOffsetPositive )
                     s->hourOffset = -s->hourOffset;
                  setRTCCOffset( s->hourOffset );
               }

               if ( !s->temporaryChanges )
               {
                  storeTimeSettingsToFile();
               }

               curHTTP.formStatus = FormStatus_OK;
               return HTTP_IO_DONE;
            }

            curHTTP.smPost = POSTState_readName;
            continue;

         default:
            break;
      }

   /* The function should never reach this point. */
   curHTTP.formStatus = FormStatus_internalErr;
   return HTTP_IO_DONE;
}
