/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "httpAuth.h"
#include <string.h>
#include <fatFS/ff.h>
#include <stdbool.h>

/* Necessary dependency chain for random and hashing functions: */
#include <TCPIPConfig.h>
#include <TCPIP Stack/StackTsk.h>
#include <GenericTypeDefs.h>
#include <TCPIP Stack/Helpers.h>
#include <TCPIP Stack/Hashes.h>

struct pwEntry
{
   unsigned long salt;
   char user[ 20 ];
   unsigned char passwordHash[ 20 ];
} static pwEntries[ HTTPAUTH_NUM_USERS ];

static void calcHash( unsigned char hash[ static 20 ], unsigned long salt,
      const char *password );
static bool authUser( int id, const char *userName, const char *password );

unsigned char HTTPNeedsAuth( unsigned char *path )
{
   if ( strncmp( (char *)&path[ 0 ], "config", 6 ) == 0 )
      return 0;
   
   return 0x80;
}

unsigned char HTTPCheckAuth( unsigned char *userName,
      unsigned char *password )
{
   int i;
   for ( i = 0; i < HTTPAUTH_NUM_USERS; ++i )
      if ( authUser( i, (const char *)userName, (const char *)password ) )
         return 0x80;

   return 0;
}

int httpAuth_loadPasswdFile()
{
   FIL passwdFile;
   FRESULT retVal;

   retVal = f_open( &passwdFile, HTTPAUTH_PASSWD_FILENAME, FA_READ );
   if ( retVal != FR_OK )
      return retVal;

   unsigned int numBytesRead;
   retVal = f_read( &passwdFile, pwEntries, sizeof( pwEntries ),
         &numBytesRead );
   if ( retVal != FR_OK )
   {
      f_close( &passwdFile );
      return retVal;
   }

   if ( numBytesRead != sizeof( pwEntries ) )
   {
      f_close( &passwdFile );
      return -1;
   }

   f_close( &passwdFile );
   return FR_OK;
}

int httpAuth_savePasswdFile()
{
   FIL passwdFile;
   FRESULT retVal;

   retVal = f_open( &passwdFile, HTTPAUTH_PASSWD_FILENAME, FA_WRITE
         | FA_CREATE_ALWAYS );
   if ( retVal != FR_OK )
      return retVal;

   unsigned int numBytesWritten;
   retVal = f_write( &passwdFile, pwEntries, sizeof( pwEntries ),
         &numBytesWritten );
   if ( retVal != FR_OK )
   {
      f_close( &passwdFile );
      return retVal;
   }

   if ( numBytesWritten != sizeof( pwEntries ) )
   {
      f_close( &passwdFile );
      return -1;
   }

   f_close( &passwdFile );

   return FR_OK;
}


int httpAuth_editUser( int id, const char *userName, const char *password )
{
   if ( id < 0 || id >= HTTPAUTH_NUM_USERS )
      return -1;

   unsigned char hash[ 20 ];
   pwEntries[ id ].salt = GenerateRandomDWORD();
   calcHash( hash, pwEntries[ id ].salt, (const char *)password );

   pwEntries[ id ].user[ 0 ] = '\0';
   strncat( pwEntries[ id ].user, userName, sizeof( pwEntries[ id ].user ) -
         1 );
   strncpy( (char *)pwEntries[ id ].passwordHash, (const char *)hash,
         sizeof( pwEntries[ id ].passwordHash ) );

   return 0;
}

static void calcHash( unsigned char hash[ static 20 ], unsigned long salt,
      const char *password )
{
   HASH_SUM shaSum;

   SHA1Initialize( &shaSum );
   HashAddROMData( &shaSum, (const unsigned char *)&salt, sizeof( salt ) );
   HashAddROMData( &shaSum, password, strlen( password ) );
   SHA1Calculate( &shaSum, hash );
}

static bool authUser( int id, const char *userName, const char *password )
{
   unsigned char hash[ 20 ];
   calcHash( hash, pwEntries[ id ].salt, password );

   return !strncmp( (const char *)pwEntries[ id ].passwordHash,
         (const char *)hash, sizeof( pwEntries[ id ].passwordHash ) );
}
