/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <xc.h>
#include "config/PICConfig.h"
#include <plib.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include "config/settings.h"
#include <TCPIP Stack/TCPIP.h>
#include "led.h"
#include "uartBuffer.h"
#include "tickTimer.h"
#include "dateTime.h"
#include "fatFS/diskio.h"

#include "calendar/calendar.h"
#include "httpAuth.h"
#include "ipACL.h"
#include "mcuC/mcuc.h"
#include "bio.h"

static void initUART();
static void initPeripherals();
static int initFileSystem();

static void enableDHCP( const char *onOff );
static void genericTestCallBack( const char *arg );
static void printNameTitle( const char *dummy );
static void checkIfMemCardPresent();

static struct command uartCallbacks[] =
{
   { "dhcp", &enableDHCP },
   { "test", &genericTestCallBack },
   { "name", &printNameTitle },
};

//void __ISR( _SPI_3_VECTOR, ipl1 ) SPI3InterruptHandler()
//{
//   INTClearFlag( INT_SOURCE_SPI_RX( 3 ) );
//   led_reverseUSR2();
//   while ( SPI3STATbits.SPIRBF )
//      _mon_putc( SPI3BUF );
//}

int main( void ) {
   SYSTEMConfig( GetSystemClock(),
         SYS_CFG_WAIT_STATES | SYS_CFG_PCACHE);
   INTEnableSystemMultiVectoredInt();

   initPeripherals();
   //mcuc_init();

   initFileSystem();
   loadSettings();

   //struct ipACLRuleSet rs =
   //{
   //   .defaultPass = false,
   //   .port = 80,
   //   .protocol = Proto_TCP,
   //};
   //StringToIPAddress( "10.10.10.10", &rs.rules[ 0 ].srcAddrRangeStart );
   //StringToIPAddress( "10.10.10.100", &rs.rules[ 0 ].srcAddrRangeEnd );
   //rs.rules[ 0 ].pass = true;
   //rs.rules[ 0 ].ruleEnabled = true;
   //ipACLSetRuleSets( &rs, 1 );

   TickInit();
   StackInit();
   puts( "Stack initialised." );

   {
      FRESULT retVal = httpAuth_loadPasswdFile();
      if ( retVal == FR_NO_FILE )
         puts( "No existing passwd file" );
      else if ( retVal )
         puts( "Could not open passwd file" );
   }

   bool IPAnnounced = false;

   for ( ;; ) {
       //mcuc_doSPIWork();
      checkIfMemCardPresent();
      uartbuf_runCmdCallbacks();
      tickTimer_handleCallbacks();

      StackTask();
      StackApplications();

      if ( !DHCPIsEnabled( 0 ) )
      {
         IPAnnounced = false;
         led_setDHCP( false );
      }

      else if ( !IPAnnounced && DHCPIsBound( 0 ) )
      {
         IPAnnounced = true;
         led_setDHCP( true );
         printf( "IP address: %d.%d.%d.%d\n",
               AppConfig.MyIPAddr.v[0],
               AppConfig.MyIPAddr.v[1],
               AppConfig.MyIPAddr.v[2],
               AppConfig.MyIPAddr.v[3]);
      }
      else if ( !DHCPIsBound( 0 ) )
      {
         led_setDHCP( false );
         IPAnnounced = false;
      }
   }
}


/* Override default putc function so that it uses the correct UART (2): */
void _mon_putc( char c )
{
   while ( U2STAbits.UTXBF );

   U2TXREG = c;
}

/* Override default exception handler so that it prints an error: */
void _general_exception_handler( unsigned cause, unsigned status )
{
   printf( "Yikes, exception!\n  Cause: %u\n  Status:  %u\n",
         ( cause & 0x1c ) >> 2, ( status & 0x7c ) >> 2 );

   for (;;);
}


static void initUART()
{
   UARTConfigure( UART2, UART_ENABLE_PINS_TX_RX_ONLY |
         UART_ENABLE_HIGH_SPEED );
   UARTSetFifoMode( UART2, UART_INTERRUPT_ON_TX_DONE |
         UART_INTERRUPT_ON_RX_NOT_EMPTY );
   UARTSetDataRate( UART2, GetPeripheralClock(), 115200 );
   UARTEnable( UART2, UART_ENABLE | UART_PERIPHERAL | UART_RX | UART_TX );

   INTSetVectorPriority( INT_VECTOR_UART( UART2 ), INT_PRIORITY_LEVEL_2 );
   INTSetVectorSubPriority( INT_VECTOR_UART( UART2 ), INT_PRIORITY_LEVEL_1 );
   INTClearFlag( INT_SOURCE_UART( UART2 ) );

   INTEnable( INT_SOURCE_UART_RX( UART2 ), INT_ENABLED );

   puts( "UART initialised." );

   uartbuf_init( uartCallbacks, sizeof( uartCallbacks ) /
         sizeof ( struct command ) );
}

static void initPeripherals()
{
   /* Make all pins digital: */
   AD1PCFG = 0xffff;

   /* Initialise LEDs: */
   led_init();

   /* Set up SD card detect pin: */
   TRISEbits.TRISE9 = 1;
   /* If card is not present, notify the file system: */
   if ( PORTEbits.RE9 )
      disk_notifyCDChange( false );

   /* Initialise RTCC: */
   initRTCC();

   /* Initialise UART: */
   initUART();
}

static int initFileSystem()
{
   int res = FileSystemInit();
   if ( !res )
      puts( "File system initialised." );
   else
      printf( "Failed to initialise file system: %d\n", res );

   return res;
}

static void enableDHCP( const char *onOff )
{
   if ( !strcmp( onOff, "on" ) )
   {
      puts( "DHCP enabled" );
      DHCPEnable( 0 );
   }

   else if ( !strcmp( onOff, "off" ) )
   {
      puts( "DHCP disabled" );
      DHCPDisable( 0 );
   }
}

static void genericTestCallBack( const char *arg )
{
   puts( "Test!" );
   //struct tm t = getRTCCDateTime();
   //printf( "Time is %02d.%02d.%02d, %02d.%02d.%02d\n",
   //      t.tm_hour, t.tm_min, t.tm_sec, t.tm_mday, t.tm_mon + 1,
   //      t.tm_year - 100 );
//   char name[ 20 ];
//   int i;
//   for ( i = 0; arg[ i + 2 ] != ' '; ++i )
//      name[ i ] = arg[ i + 2 ];
//   name[ i ] = '\0';
//   printf( "Name: %s\nPassword: %s\n", name, &arg[ i + 3 ] );
//
//   httpAuth_editUser( arg[ 0 ] - '0', name, &arg[ i + 3 ] );
//   FRESULT retVal = httpAuth_savePasswdFile();
//   if ( retVal )
//      printf( "Could not save passwd file: %d\n", retVal );
}

static void printNameTitle( const char *dummy )
{
   char *name = bio_name();
   if ( !name )
      puts( "Name error" );
   else
   {
      printf( "Name: '%s'\n", name );
      free( name );
   }

   char *title = bio_title();
   if ( !title )
      puts( "Title error" );
   else
   {
      printf( "Title: '%s'\n", title );
      free( title );
   }
}

static void checkIfMemCardPresent()
{
   static bool stateEverProbed = false;
   static bool prevState = true;
   static bool waitingForDebounce = false;
   static DWORD prevTick = 0;

   if ( !stateEverProbed )
   {
      stateEverProbed = true;
      prevState = PORTEbits.RE9;
      printf( "SD card%s present\n", !prevState ? "" : " not" );
      return;
   }

   /* Wait for switch debouncing for 100 ms before determining final state: */
   if ( waitingForDebounce )
   {
      if ( TickGet() - prevTick > TICKS_PER_SECOND / 10 )
      {
         waitingForDebounce = false;

         if ( PORTEbits.RE9 == prevState )
            return;

         prevState = PORTEbits.RE9;
         disk_notifyCDChange( !prevState );
         HTTPReset();
         printf( "SD card%s present\n", !prevState ? "" : " not" );
      }
   }

   else if ( PORTEbits.RE9 != prevState )
   {
      prevTick = TickGet();
      waitingForDebounce = true;
   }
}
