/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "tickTimer.h"
#include "HardwareProfile.h"
#include <TCPIP Stack/Tick.h>

/**
 * Callback, with function, argument and information about when to fire
 */
struct callback
{
   /**
    * Callback function, returning nothing, taking one argeument, a void
    * pointer
    */
   void ( *func )( void * );
   /** Argument that will be passed to func() */
   void *arg;
   /**
    * Number of times the callbacks should be called (including the first
    * call)
    */
   int count;
   /** The last time the callback fired. Initially 0. */
   unsigned int lastFire;
   /** The interval the callback should be called (in ms) */
   unsigned int intervalMs;
} static callbacks[ TICKTIMER_MAX_CALLBACKS ];


int tickTimer_addCallback( void ( *callback )( void *arg ), void *arg,
      int intervalMSecs, int repeats )
{
   int i;
   struct callback *vacantCallBack = NULL;

   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
      if ( !callbacks[ i ].func )
      {
         vacantCallBack = &callbacks[ i ];
         break;
      }

   if ( !vacantCallBack )
      return 1;

   vacantCallBack->func = callback;
   vacantCallBack->arg = arg;
   vacantCallBack->count = 1 + repeats;
   vacantCallBack->intervalMs = intervalMSecs;
   vacantCallBack->lastFire = TickGet();

   return 0;
}

void tickTimer_removeCallback( void ( *callback )( void *arg ) )
{
   int i;
   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
      if ( callbacks[ i ].func == callback )
      {
         callbacks[ i ].func = NULL;
         return;
      }
}

bool tickTimer_callbackActive( void ( *callback )( void *arg ) )
{
   int i;
   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
      if ( callbacks[ i ].func == callback )
         return true;

   return false;
}

void tickTimer_handleCallbacks()
{
   unsigned int currTick = TickGet();

   int i;
   for ( i = 0; i < TICKTIMER_MAX_CALLBACKS; ++i )
   {
      if ( callbacks[ i ].func && ( currTick - callbacks[ i ].lastFire ) >=
            TICKS_PER_SECOND / 1000 * callbacks[ i ].intervalMs )
      {
         ( *callbacks[ i ].func)( callbacks[ i ].arg );
         callbacks[ i ].lastFire = currTick;

         --callbacks[ i ].count;
         if ( !callbacks[ i ].count )
            callbacks[ i ].func = NULL;
      }
   }
}
