/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \addtogroup TickTimer
 * @{
 *
 * \file
 */

#ifndef TICKTIMER_H
#define TICKTIMER_H

#include <stdbool.h>

/**
 * \brief Size of the callback function array
 */
#define TICKTIMER_MAX_CALLBACKS 15

/**
 * \brief Add a callback function
 *
 * The provided function will be called with the given argument after the
 * given interval has passed. It is then removed from the callback array,
 * providing a vacant space for a new callback function. If \pn{repeats} is
 * non-zero, the function will be called repeatedly with the given interval
 * before it is removed. If \pn{repeats} is a negative number, the function
 * will be repeated forever (and forever occupy space in the callback array).
 *
 * Note that callbacks are called in the order they are added.
 *
 * \param callback a pointer to a function returning nothing, expecting one
 * void pointer argument
 * \param arg the argument which will be passed to the \pn{ callback }
 * function. May be NULL
 * \param intervalMSecs interval in milliseconds
 * \param repeats number of times the function will be repeated. If 0, the
 * function will be called only once. If any negative value, the function will
 * be repeated forever.
 *
 * \retval 0 if the callback was registered successfully, non-zero if there
 * was no more room in the interal callback array
 */
int tickTimer_addCallback( void ( *callback )( void *arg ), void *arg,
      int intervalMSecs, int repeats );

/**
 * \brief Remove a previously registered callback
 *
 * \param callback a pointer to the callback function to remove/unregister
 */
void tickTimer_removeCallback( void ( *callback )( void *arg ) );

/**
 * \brief Check if callback is active
 *
 * Active here means that \pn{callback} is in the callback array.
 *
 * \param callback a pointer to the callback function to check
 *
 * \retval true if the callback exists in the callback array
 * \retval false if the callback is not currently active
 */
bool tickTimer_callbackActive( void ( *callback )( void *arg ) );

/**
 * \brief Check all registered callbacks and see if they chould be called
 * given the current tick value
 *
 * This functions must be called often enough in order to ensure a timely
 * response. Note that callbacks are called in the order they are added.
 */
void tickTimer_handleCallbacks();

#endif // TICKTIMER_H

/**
 * @}
 */
