/* 
 * Copyright (C) 2013 Andreas Misje
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "uartBuffer.h"
#ifdef __PIC24F__
#include "peripheralLibConfig.h"
#include <uart.h>
#else
#include <plib.h>
#endif // __PIC24F__

#define PASTE( a, b ) a ## b
#define CONCAT( a,b ) PASTE( a, b )

#ifdef __PIC24F__
#define enableRXInterrupt() \
   ( CONCAT( EnableIntU, CONCAT( UARTBUF_UART_MODULE, RX ) ) )
#define disableRXInterrupt() \
   ( CONCAT( DisableIntU, CONCAT( UARTBUF_UART_MODULE, RX ) ) )
#define clearRXInterruptFlag() \
   ( CONCAT( U, CONCAT( UARTBUF_UART_MODULE, RX_Clear_Intr_Status_Bit ) ) )
#else
#define enableRXInterrupt() \
   INTEnable( CONCAT( INT_U, CONCAT( UARTBUF_UART_MODULE, RX ) ), \
         INT_ENABLED )
#define disableRXInterrupt() \
   INTEnable( CONCAT( INT_U, CONCAT( UARTBUF_UART_MODULE, RX ) ), \
         INT_DISABLED )
#define clearRXInterruptFlag() \
   INTClearFlag( CONCAT( INT_U, CONCAT( UARTBUF_UART_MODULE, RX ) ) )
#endif // __PIC24F__

#define URXREG ( CONCAT( U, CONCAT( UARTBUF_UART_MODULE, RXREG ) ) )

struct ringBuffer
{
   char buffer[ UARTBUF_RX_RINGBUFFER_SIZE ];
   size_t start;
   size_t count;
} volatile static ringBuffer;

#define ringBufIncr( i ) ( i = ( i + 1 ) % UARTBUF_RX_RINGBUFFER_SIZE )
#define ARG_LEN ( UARTBUF_CMD_ARG_LEN + 1 )

static volatile size_t prevNewlinePos;
static volatile size_t newlinePos;
static volatile bool checkCommands;

struct command *commandList;
size_t commandListCount;

static bool ringBufStrExactCmp( const char *string2 );

void uartbuf_init( struct command *commands, size_t numCommands )
{
   ringBuffer.start = 0;
   ringBuffer.count = 0;

   checkCommands = false;
   prevNewlinePos = -1;
   newlinePos = -1;
   commandList = commands;
   commandListCount = numCommands;
}

void uartbuf_runCmdCallbacks()
{
   if ( !commandList )
      return;

   if ( !checkCommands )
      return;

   disableRXInterrupt();

   checkCommands = false;

   char ch;
   size_t i, spacePos;
   for ( i = ( prevNewlinePos + 1 ) % UARTBUF_RX_RINGBUFFER_SIZE; ;
         ringBufIncr( i ) )
      if ( ringBuffer.buffer[ i ] == ' ' || ringBuffer.buffer[ i ] == '\n' )
         break;

   spacePos = i;
   ringBuffer.buffer[ i ] = '\0';

   for ( i = 0; i < commandListCount; ++i )
      if ( commandList[ i ].command && ringBufStrExactCmp( commandList[ i ]
               .command ) )
         break;

   if ( i == commandListCount )
   {
      enableRXInterrupt();
      return;
   }

   void (* callback)( const char *arg ) = commandList[ i ].callback;

   char arg[ ARG_LEN ] = "";

   if ( spacePos != newlinePos )
   {
      for ( i = 0; i < ARG_LEN; ++i )
      {
         ch = ringBuffer.buffer[ ( spacePos + 1 + i ) %
            UARTBUF_RX_RINGBUFFER_SIZE ];
         if ( ch == '\n' )
            break;

         arg[ i ] = ch;
      }
      if ( i == ARG_LEN )
      {
         enableRXInterrupt();
         return;
      }

      arg[ i ] = '\0';
   }

   enableRXInterrupt();

   callback( arg );
}


static bool ringBufStrExactCmp( const char *string2 )
{
   size_t i;
   for ( i = 0; ; ++i )
   {
      char ch = ringBuffer.buffer[ ( prevNewlinePos + 1 + i ) %
         UARTBUF_RX_RINGBUFFER_SIZE ];

      if ( ch && string2[ i ] )
      {
         if ( ch != string2[ i ] )
            return false;
      }
      else if ( !ch && !string2[ i ] )
         return true;
      else
         return false;
   }

   return true;
}


size_t uartbuf_copyBuffer( char *buffer, size_t len )
{
   if ( !buffer )
      return 0;

   size_t i;
   for ( i = 0; i < len && i < ringBuffer.count; ++i )
      buffer[ i ] = ringBuffer.buffer[ ( ringBuffer.start + i ) %
         UARTBUF_RX_RINGBUFFER_SIZE ];

   return i;
}


#ifdef __PIC24F__
void __attribute__(( interrupt, auto_psv )) CONCAT( _U, CONCAT( \
         UARTBUF_UART_MODULE, RXInterrupt() ) )
#else
void __ISR( CONCAT( _UART, CONCAT( UARTBUF_UART_MODULE, _VECTOR ) ),
      CONCAT( ipl, UARTBUF_UART_INT_PRI ) )
      uartRingBufferInterruptHandler( void )
#endif
{
   size_t end = ( ringBuffer.start + ringBuffer.count ) %
      UARTBUF_RX_RINGBUFFER_SIZE;
   char data = URXREG;

   ringBuffer.buffer[ end ] = data;

   if ( ringBuffer.count == UARTBUF_RX_RINGBUFFER_SIZE )
      ringBufIncr( ringBuffer.start );
   else
      ++ringBuffer.count;

   if ( data == '\n' )
   {
      prevNewlinePos = newlinePos;
      newlinePos = end;
      checkCommands = true;
   }

   clearRXInterruptFlag();
}
